""" This module contains functions to read spectra and save deltas. Can be
imported without a need for MPI."""

import argparse
import functools
import warnings

import fitsio
import numpy as np
from numpy.lib.recfunctions import drop_fields, merge_arrays

from qsonic import QsonicException
import qsonic.spectrum


def add_io_parser(parser=None):
    """ Adds IO related arguments to parser. These arguments are grouped under
    'Input/output parameters and selections'. ``--input-dir`` and ``--catalog``
    are required.

    Arguments
    ---------
    parser: argparse.ArgumentParser, default: None

    Returns
    ---------
    parser: argparse.ArgumentParser
    """
    if parser is None:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    ingroup = parser.add_argument_group('Input options')
    ingroup.add_argument(
        "--input-dir", '-i', required=True,
        help="Input directory.")
    ingroup.add_argument(
        "--catalog", required=True,
        help="Catalog filename")
    ingroup.add_argument(
        "--tile-format", action="store_true",
        help="Read tile coadd-*.fits files in tiles/cumulative directory.")
    ingroup.add_argument(
        "--mock-analysis", action="store_true",
        help="Input folder is mock. Uses nside=16")
    ingroup.add_argument(
        "--keep-surveys", nargs='+', default=['main'],
        help="Surveys to keep.")
    ingroup.add_argument(
        "--skip-resomat", action="store_true",
        help="Skip reading resolution matrix for 3D.")
    ingroup.add_argument(
        "--arms", default=['B', 'R'], choices=['B', 'R', 'Z'], nargs='+',
        help="Arms to read.")

    outgroup = parser.add_argument_group('Output options')
    outgroup.add_argument(
        "--outdir", '-o',
        help="Output directory to save deltas.")
    outgroup.add_argument(
        "--coadd-arms", default="before",
        choices=["before", "after", "disable"],
        help="Coadds arms before or after continuum fitting or not at all.")
    outgroup.add_argument(
        "--exposures", default="disable",
        choices=["before", "after", "disable"],
        help=("Reads exposures before or after continuum fitting and saves. "
              "Tile format not supported. Related function "
              ":func:`qsonic.scripts.qsonic_fit.mpi_read_exposures_after`."
              ))
    outgroup.add_argument(
        "--save-by-hpx", action="store_true",
        help="Save by healpix. If not, saves by MPI rank.")
    return parser


def get_spectra_reader_function(
        input_dir, arms_to_keep, mock_analysis, skip_resomat,
        read_true_continuum, is_tile, read_exposures, program="dark"
):
    """ Returns a callable object (function) that returns a list of Spectrum
    objects for a given catalog of a single healpix. Essentially, a wrapper
    for the following functions:

        - :meth:`read_onehealpix_file_mock`,
        - :meth:`read_onetile_coaddfile_data`,
        - :meth:`read_onehealpix_file_data_uncoadd`,
        - :meth:`read_onehealpix_file_data_coadd`.

    If for data, ``SURVEY`` column must be present and sorted when calling the
    function this returns in the healpix grouping. Tile grouping requires
    ``TILEID`` and ``PETAL_LOC``.

    Arguments
    ---------
    input_dir: str
        Input directory
    arms_to_keep: list(str)
        Must only contain B, R and Z
    mock_analysis: bool
        Reads for mock data if true.
    skip_resomat: bool
        If true, do not read resomat.
    read_true_continuum: bool
        If true, reads the true continuum for mock analysis.
    is_tile: bool
        If true, reads for tile grouping
    read_exposures: bool
        If true, reads exposures for healpix format only.
    program: str, default: "dark"
        Always use dark program.

    Returns
    ---------
    readerFunction: Callable
        Call this function with a catalog of quasars in a single healpix.

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    if mock_analysis:
        return functools.partial(
            read_onehealpix_file_mock,
            input_dir=input_dir, arms_to_keep=arms_to_keep,
            skip_resomat=skip_resomat, read_true_continuum=read_true_continuum
        )

    elif is_tile:
        return functools.partial(
            read_onetile_coaddfile_data,
            input_dir=input_dir, arms_to_keep=arms_to_keep,
            skip_resomat=skip_resomat
        )

    elif read_exposures:
        return functools.partial(
            read_onehealpix_file_data_uncoadd,
            input_dir=input_dir, arms_to_keep=arms_to_keep,
            skip_resomat=skip_resomat, program=program
        )

    else:
        return functools.partial(
            read_onehealpix_file_data_coadd,
            input_dir=input_dir, arms_to_keep=arms_to_keep,
            skip_resomat=skip_resomat, program=program
        )


def read_resolution_matrices_onehealpix_data(
        catalog_hpx, input_dir, spectra_list, program="dark"
):
    """ Returns a list of Spectrum objects for a given catalog of a single
    healpix.

    Arguments
    ---------
    catalog_hpx: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog of quasars in a single healpix. If for data 'SURVEY' column
        must be present and sorted.
    input_dir: str
        Input directory
    spectra_list: list(Spectrum)
        List of spectra to read resolution matrix
    program: str, default: "dark"
        Always use dark program.

    Returns
    ---------
    spectra_list: list(Spectrum)

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    # assert (catalog_hpx.size == len(spectra_list))

    unique_surveys, s2 = np.unique(catalog_hpx['SURVEY'], return_index=True)
    survey_split_cat = np.split(catalog_hpx, s2[1:])
    s2 = np.append(s2, len(spectra_list))

    arms_to_keep = spectra_list[0].wave.keys()

    for ii, cat_by_survey in enumerate(survey_split_cat):
        survey = cat_by_survey['SURVEY'][0]
        pixnum = cat_by_survey['HPXPIXEL'][0]
        targetids_by_survey = cat_by_survey['TARGETID']

        fspec = (f"{input_dir}/{survey}/{program}/{pixnum//100}/"
                 f"{pixnum}/coadd-{survey}-{program}-{pixnum}.fits")

        i1, i2 = s2[ii], s2[ii + 1]
        spectra_list[i1:i2] = _read_onehealpix_file_onlyreso(
            targetids_by_survey, fspec, arms_to_keep, spectra_list[i1:i2])

    return spectra_list


def read_deltas(fname):
    """ Returns a list of all Delta objects in a file.

    Arguments
    ---------
    fname: str
        FITS file name

    Returns
    ---------
    deltas_list: list(Delta)

    Raises
    ---------
    RuntimeError
        If the file is missing certain columns and keys. See :class:`Delta`.
    """

    with fitsio.FITS(fname) as fts:
        deltas_list = [qsonic.spectrum.Delta(hdu) for hdu in fts[1:]]

    return deltas_list


def save_deltas(
        spectra_list, outdir, save_by_hpx=False, mpi_rank=None
):
    """ Saves given list of spectra as deltas. NO coaddition of arms.
    Each arm is saved separately. Only valid spectra are saved.

    Arguments
    ---------
    spectra_list: list(Spectrum)
        Continuum fitted spectra objects. All must be valid!
    outdir: str
        Output directory. Does not save if empty of None
    save_by_hpx: bool, default: False
        Saves by healpix if True. Has priority over mpi_rank
    mpi_rank: int, default: None
        Rank of the MPI process. Save by `mpi_rank` if passed.

    Raises
    ---------
    QsonicException
        If both `mpi_rank` and `save_by_hpx` is None/False.
    QsonicException
        If blinding is not set.
    """
    if not outdir:
        return

    if save_by_hpx:
        pixnos = np.array([spec.hpix for spec in spectra_list])
        sort_idx = np.argsort(pixnos)
        pixnos = pixnos[sort_idx]
        unique_pix, s = np.unique(pixnos, return_index=True)
        split_spectra = np.split(np.array(spectra_list)[sort_idx], s[1:])
    elif mpi_rank is not None:
        unique_pix = [mpi_rank]
        split_spectra = [spectra_list]
    else:
        raise QsonicException("save_by_hpx and mpi_rank can't both be None.")

    if qsonic.spectrum.Spectrum.blinding_not_set():
        raise QsonicException("Blinding is not set. Cannot save delta.")

    for healpix, hp_specs in zip(unique_pix, split_spectra):
        results = fitsio.FITS(
            f"{outdir}/delta-{healpix}.fits", 'rw', clobber=True)

        for spec in qsonic.spectrum.valid_spectra(hp_specs):
            spec.write(results)

        results.close()


def _read_resoimage(imhdu, quasar_indices, nwave):
    # Reading into sorted order is faster
    # Since the output catalog is already sorted, we place them in order
    ndiags = imhdu.get_dims()[1]
    dtype = imhdu._get_image_numpy_dtype()
    data = np.empty((quasar_indices.size, ndiags, nwave), dtype=dtype)
    for idata, iqso in enumerate(quasar_indices):
        data[idata, :, :] = imhdu[int(iqso), :, :]

    return data


def _read_imagehdu(imhdu, quasar_indices, nwave):
    dtype = imhdu._get_image_numpy_dtype()
    data = np.empty((quasar_indices.size, nwave), dtype=dtype)
    for idata, iqso in enumerate(quasar_indices):
        data[idata, :] = imhdu[int(iqso), :]

    return data

# Timing showed this is slower
# def _read_imagehdu_2(imhdu, quasar_indices):
#     ndims = imhdu.get_info()['ndims']
#     if ndims == 2:
#         return np.vstack([imhdu[int(idx), :] for idx in quasar_indices])
#     return np.vstack([imhdu[int(idx), :, :] for idx in quasar_indices])


def _read_true_continuum(targetids, fspec):
    """Read true continuum as dictionary from file.

    Arguments
    ---------
    targetids: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Target IDs.
    fspec: str
        Filename to open.

    Returns
    ---------
    cont_data: dict( :external+numpy:py:class:`ndarray <numpy.ndarray>` )
        This has keys for w1, w2, dwave and data, where data is ndarray and
        others are floats.

    Raises
    ---------
    RuntimeError
        If number of quasars in TRUE_CONT does not match the input.
    """
    with fitsio.FITS(fspec) as fitsfile:
        hdr = fitsfile['TRUE_CONT'].read_header()
        true_continua = fitsfile['TRUE_CONT'].read()
    w1 = hdr['WMIN']
    w2 = hdr['WMAX']
    dw = hdr['DWAVE']

    common_targetids, idx_fbr, _ = np.intersect1d(
        true_continua['TARGETID'], targetids,
        assume_unique=True, return_indices=True
    )

    if (common_targetids.size != targetids.size):
        raise RuntimeError(
            f"Error reading true continua from {fspec}. "
            "Number of quasars in TRUE_CONT does not match the catalog "
            f"catalog:{targetids.size} vs "
            f"healpix:{common_targetids.size}!")

    cont_data = {
        'w1': w1, 'w2': w2, 'dwave': dw,
        'data': true_continua['TRUE_CONT'][idx_fbr, :].astype("f8")
    }

    return cont_data


def _read_onehealpix_file(
        targetids_by_survey, fspec, arms_to_keep, skip_resomat,
        fbrmap_columns=['TARGETID']
):
    """Common function to read a single FITS file in DESI healpix grouping.

    Arguments
    ---------
    targetids_by_survey: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Targetids_by_survey (used to be catalog). If data, split by survey and
        contains only one survey.
    fspec: str
        Filename to open.
    arms_to_keep: list(str)
        Must only contain B, R and Z.
    skip_resomat: bool
        If true, do not read resomat.
    fbrmap_columns: list(str), default: ['TARGETID']
        Columns to read from FIBERMAP. For coadded files, it should be
        ``['TARGETID']``. For uncoadded files (spectra-), you should read at
        least ``[TARGETID, PETAL_LOC, FIBER, NIGHT, EXPID, TILEID]``.

    Returns
    ---------
    data: dict( :external+numpy:py:class:`ndarray <numpy.ndarray>` )
        Only quasar spectra are read into keywords wave, flux etc. Resolution
        is read if present.
    idx_cat: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Indices in `targetids_by_survey` there were succesfully read.
    fbrmap: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog with ``columns`` from the FIBERMAP. Note the size will not be
        equal to ``targetids_by_survey.size`` when reading exposures.

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    fitsfile = fitsio.FITS(fspec)

    fbrmap = fitsfile['FIBERMAP'].read(columns=fbrmap_columns)

    common_targetids, _, idx_cat = np.intersect1d(
        fbrmap['TARGETID'], targetids_by_survey, return_indices=True)
    if (common_targetids.size != targetids_by_survey.size):
        warnings.warn(
            f"Error reading {fspec}. "
            "Number of quasars in healpix does not match the catalog "
            f"catalog:{targetids_by_survey.size} vs "
            f"healpix:{common_targetids.size}!", RuntimeWarning)

    idx_fbr = np.nonzero(np.isin(fbrmap['TARGETID'], targetids_by_survey))[0]
    targetids_fbr = fbrmap['TARGETID'][idx_fbr]
    idx_fbr = idx_fbr[targetids_fbr.argsort()]

    data = {
        'wave': {},
        'flux': {},
        'ivar': {},
        'mask': {},
        'reso': {}
    }

    for arm in arms_to_keep:
        # Cannot read by rows= argument.
        data['wave'][arm] = fitsfile[f'{arm}_WAVELENGTH'].read()
        nwave = data['wave'][arm].size

        data['flux'][arm] = _read_imagehdu(
            fitsfile[f'{arm}_FLUX'], idx_fbr, nwave)
        data['ivar'][arm] = _read_imagehdu(
            fitsfile[f'{arm}_IVAR'], idx_fbr, nwave)
        data['mask'][arm] = _read_imagehdu(
            fitsfile[f'{arm}_MASK'], idx_fbr, nwave)

        if skip_resomat or f'{arm}_RESOLUTION' not in fitsfile:
            continue

        data['reso'][arm] = _read_resoimage(
            fitsfile[f'{arm}_RESOLUTION'], idx_fbr, nwave)

    fitsfile.close()

    return data, idx_cat, fbrmap[idx_fbr]


def _read_onehealpix_file_onlyreso(
        targetids_by_survey, fspec, arms_to_keep, spectra_list
):
    """ Function to read forest region resolution matrix for data. Note reading
    resolution matrix is already fast for mocks.

    Arguments
    ---------
    targetids_by_survey: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Targetids_by_survey (used to be catalog). If data, split by survey and
        contains only one survey.
    fspec: str
        Filename to open.
    arms_to_keep: list(str)
        Must only contain B, R and Z.
    spectra_list: list(Spectrum)
        List of spectra that forest region is set. Modified in place and also
        returned.

    Returns
    ---------
    spectra_list: list(Spectrum)
        List of spectra where forestreso is set.

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    fitsfile = fitsio.FITS(fspec)

    fbrmap = fitsfile['FIBERMAP'].read(columns='TARGETID')
    common_targetids, idx_fbr, idx_cat = np.intersect1d(
        fbrmap, targetids_by_survey, assume_unique=True, return_indices=True)
    if (common_targetids.size != targetids_by_survey.size):
        warnings.warn(
            f"Error reading {fspec}. "
            "Number of quasars in healpix does not match the catalog "
            f"catalog:{targetids_by_survey.size} vs "
            f"healpix:{common_targetids.size}!", RuntimeWarning)

    for arm in arms_to_keep:
        imhdu = fitsfile[f'{arm}_RESOLUTION']

        for jj, spec in zip(idx_fbr, spectra_list):
            # assert (common_targetids[jj] == spec.targetid)

            if arm not in spec.forestwave.keys():
                continue

            f1 = spec._f1[arm]
            f2 = spec._f2[arm]
            spec._forestreso[arm] = imhdu[int(jj), :, f1:f2][0]

    fitsfile.close()

    return spectra_list


def read_onehealpix_file_data_coadd(
        catalog_hpx, input_dir, arms_to_keep, skip_resomat, program="dark"
):
    """Read FITS files for all surveys needed for data.

    Arguments
    ---------
    catalog_hpx: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog for a healpix. Ordered by SURVEY and TARGETID.
    input_dir: str
        Input directory.
    arms_to_keep: list(str)
        Must only contain B, R and Z.
    skip_resomat: bool
        If true, do not read resomat.
    program: str, default: "dark"
        Program.

    Returns
    ---------
    spectra_list: list(Spectrum)

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    spectra_list = []

    survey_split_cat = np.split(
        catalog_hpx, np.unique(catalog_hpx['SURVEY'], return_index=True)[1][1:]
    )
    pixnum = catalog_hpx['HPXPIXEL'][0]

    for cat_by_survey in survey_split_cat:
        survey = cat_by_survey['SURVEY'][0]

        fspec = (f"{input_dir}/{survey}/{program}/{pixnum//100}/"
                 f"{pixnum}/coadd-{survey}-{program}-{pixnum}.fits")
        data, idx_cat, _ = _read_onehealpix_file(
            cat_by_survey['TARGETID'], fspec, arms_to_keep, skip_resomat)

        if idx_cat.size != cat_by_survey.size:
            cat_by_survey = cat_by_survey[idx_cat]

        spectra_list.extend(
            qsonic.spectrum.generate_spectra_list_from_data(
                cat_by_survey, data)
        )

    return spectra_list


def read_onehealpix_file_data_uncoadd(
        catalog_hpx, input_dir, arms_to_keep, skip_resomat, program="dark"
):
    """Read uncoadded spectra from FITS files for all surveys needed for data.

    Arguments
    ---------
    catalog_hpx: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog for a healpix. Ordered by SURVEY and TARGETID.
    input_dir: str
        Input directory.
    arms_to_keep: list(str)
        Must only contain B, R and Z.
    skip_resomat: bool
        If true, do not read resomat.
    program: str, default: "dark"
        Program.

    Returns
    ---------
    spectra_list: list(Spectrum)
    """
    spectra_list = []

    survey_split_cat = np.split(
        catalog_hpx, np.unique(catalog_hpx['SURVEY'], return_index=True)[1][1:]
    )
    pixnum = catalog_hpx['HPXPIXEL'][0]

    for cat_by_survey in survey_split_cat:
        survey = cat_by_survey['SURVEY'][0]

        fspec = (f"{input_dir}/{survey}/{program}/{pixnum//100}/"
                 f"{pixnum}/spectra-{survey}-{program}-{pixnum}.fits")

        data, idx_cat, fbrmap = _read_onehealpix_file(
            cat_by_survey['TARGETID'], fspec, arms_to_keep, skip_resomat,
            ['TARGETID', 'PETAL_LOC', 'FIBER', 'NIGHT', 'EXPID', 'TILEID'])

        if idx_cat.size != cat_by_survey.size:
            cat_by_survey = cat_by_survey[idx_cat]

        new_cat = cat_by_survey[
            np.searchsorted(cat_by_survey['TARGETID'], fbrmap['TARGETID'])]
        new_cat = merge_arrays(
            (new_cat, drop_fields(fbrmap, 'TARGETID', usemask=False)),
            flatten=True, usemask=False
        )

        spectra_list.extend(
            qsonic.spectrum.generate_spectra_list_from_data(new_cat, data)
        )

    return spectra_list


def read_onetile_coaddfile_data(
        catalog_tile, input_dir, arms_to_keep, skip_resomat
):
    """Read all petal coadd FITS files for a given tile.

    Arguments
    ---------
    catalog_tile: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog for a tile. Ordered by PETAL_LOC and TARGETID.
    input_dir: str
        Input directory.
    arms_to_keep: list(str)
        Must only contain B, R and Z.
    skip_resomat: bool
        If true, do not read resomat.

    Returns
    ---------
    spectra_list: list(Spectrum)

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the tile file does not match the catalog.
    """
    tileid = catalog_tile['TILEID'][0]
    lastnight = catalog_tile['LASTNIGHT'][0]

    petal_split_cat = np.split(
        catalog_tile, np.unique(
            catalog_tile['PETAL_LOC'], return_index=True)[1][1:]
    )

    spectra_list = []
    for cat_by_petal in petal_split_cat:
        petal = cat_by_petal['PETAL_LOC'][0]

        fspec = (f"{input_dir}/{tileid}/{lastnight}/"
                 f"coadd-{petal}-{tileid}-thru{lastnight}.fits")
        data, idx_cat, _ = _read_onehealpix_file(
            cat_by_petal['TARGETID'], fspec, arms_to_keep, skip_resomat)

        if idx_cat.size != cat_by_petal.size:
            cat_by_petal = cat_by_petal[idx_cat]

        spectra_list.extend(
            qsonic.spectrum.generate_spectra_list_from_data(
                cat_by_petal, data)
        )

    return spectra_list


def read_onehealpix_file_mock(
        catalog_hpx, input_dir, arms_to_keep, skip_resomat,
        read_true_continuum, nside=16
):
    """ Read a single FITS file for mocks.

    Arguments
    ---------
    catalog_hpx: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog for a single healpix. Ordered by TARGETID.
    input_dir: str
        Input directory.
    arms_to_keep: list(str)
        Must only contain B, R and Z.
    skip_resomat: bool
        If true, do not read resomat.
    read_true_continuum: bool
        If true, reads the true continuum for mock analysis.
    nside: int, default: 16
        NSIDE for healpix.

    Returns
    ---------
    spectra_list: list(Spectrum)

    Raises
    ---------
    RuntimeWarning
        If number of quasars in the healpix file does not match the catalog.
    """
    pixnum = catalog_hpx['HPXPIXEL'][0]
    fspec = f"{input_dir}/{pixnum//100}/{pixnum}/spectra-{nside}-{pixnum}.fits"
    data, idx_cat, _ = _read_onehealpix_file(
        catalog_hpx['TARGETID'], fspec, arms_to_keep, skip_resomat)

    if idx_cat.size != catalog_hpx.size:
        catalog_hpx = catalog_hpx[idx_cat]

    fspec = f"{input_dir}/{pixnum//100}/{pixnum}/truth-{nside}-{pixnum}.fits"
    if read_true_continuum:
        data['cont'] = _read_true_continuum(catalog_hpx['TARGETID'], fspec)

    if skip_resomat:
        return qsonic.spectrum.generate_spectra_list_from_data(
            catalog_hpx, data)

    with fitsio.FITS(fspec) as fitsfile:
        for arm in arms_to_keep:
            data['reso'][arm] = np.array(fitsfile[f'{arm}_RESOLUTION'].read())

    return qsonic.spectrum.generate_spectra_list_from_data(catalog_hpx, data)


def _float_range(f1, f2):
    # Define the function with default arguments
    def float_range_checker(arg):
        """New Type function for argparse - a float within predefined range.
        """
        try:
            f = float(arg)
        except ValueError:
            raise argparse.ArgumentTypeError("must be a floating point number")
        if f < f1 or f > f2:
            raise argparse.ArgumentTypeError(f"must be in range [{f1}--{f2}]")
        return f

    # Return function handle to checking function
    return float_range_checker



""" Calibration module for noise and flux. """
import argparse

import fitsio
import numpy as np

from qsonic.mathtools import (
    FastCubic1DInterp, FastLinear1DInterp, _one_function)
from qsonic.mpi_utils import mpi_fnc_bcast


def add_calibration_parser(parser=None):
    """ Adds calibration related arguments to parser. These arguments are
    grouped under 'Noise and flux calibation options'.

    Arguments
    ---------
    parser: argparse.ArgumentParser, default: None

    Returns
    ---------
    parser: argparse.ArgumentParser
    """
    if parser is None:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    calib_group = parser.add_argument_group(
        'Noise and flux calibation options')

    calib_group.add_argument(
        "--noise-calibration", help="Noise calibration file.")
    calib_group.add_argument(
        "--varlss-as-additive-noise", action="store_true",
        help="var_lss as additive noise term after continuum fitting.")
    calib_group.add_argument(
        "--flux-calibration", help="Flux calibration file.")

    return parser


class NoiseCalibrator():
    """ Noise calibration object.

    .. math:: i \\rightarrow i / \\eta,

    where i is IVAR.

    FITS file must have 'VAR_FUNC' extension. This extension must have columns
    for 'lambda' and 'eta'. Wavelength array must be linearly and equally
    spaced. Uses cubic spline.

    Parameters
    ----------
    fname: str
        Filename to read by ``fitsio``.
    comm: None or MPI.COMM_WORLD, default: None
    mpi_rank: int, default: 0
    add_varlss: bool, default: False
        Use var_lss as an additive correction to noise in observed frame.
        Requires continuum.
    no_eta: bool, default: False
        Turns off eta scaling by using eta=1.

    Attributes
    ----------
    eta_interp: FastCubic1DInterp
        Eta interpolator.
    """

    def _read(self, fname):
        with fitsio.FITS(fname) as fts:
            data = fts['VAR_FUNC'].read()

        waves = data['lambda']
        waves_0 = waves[0]
        dwave = waves[1] - waves[0]

        if not np.allclose(np.diff(waves), dwave):
            raise Exception(
                "Failed to construct noise calibration from "
                f"{fname}::wave is not equally spaced.")

        var_lss = np.array(data['var_lss'], dtype='d')
        varlss_interp = FastCubic1DInterp(waves_0, dwave, var_lss)

        eta = np.array(data['eta'], dtype='d')
        eta[eta == 0] = 1
        eta_interp = FastCubic1DInterp(waves_0, dwave, eta)

        return eta_interp, varlss_interp

    def __init__(
            self, fname, comm=None, mpi_rank=0,
            add_varlss=False, no_eta=False
    ):
        self.eta_interp, self.varlss_interp = mpi_fnc_bcast(
            self._read, comm, mpi_rank,
            f"Error loading NoiseCalibrator from file {fname}.",
            fname)

        if not add_varlss:
            self.varlss_interp = None

        if no_eta:
            self.eta_interp = _one_function

    def apply(self, spectra_list):
        """ Apply the noise calibration by **only** scaling
        :attr:`forestivar <qsonic.spectrum.Spectrum.forestivar>`. Smooth
        component must be set after this.

        Arguments
        ----------
        spectra_list: list(Spectrum)
            Spectrum objects to noise calibrate.
        """
        if self.varlss_interp:
            if not all(spec.cont_params['cont'] for spec in spectra_list):
                raise Exception(
                    "NoiseCalibrator needs continuum for additive correction.")

            for spec in spectra_list:
                for arm, wave_arm in spec.forestwave.items():
                    eta = self.eta_interp(wave_arm)
                    vlss = self.varlss_interp(wave_arm)
                    vlss *= spec.cont_params['cont'][arm]**2
                    spec.forestivar[arm] /= (eta + spec.forestivar[arm] * vlss)
                    spec.forestivar[arm][spec.forestivar[arm] < 0] = 0
        else:
            for spec in spectra_list:
                for arm, wave_arm in spec.forestwave.items():
                    eta = self.eta_interp(wave_arm)
                    spec.forestivar[arm] /= eta


class FluxCalibrator():
    """ Flux calibration object.

    .. math::

        f &\\rightarrow f / s

        i &\\rightarrow i \\times s^2,

    where i is IVAR and s is the stacked flux.

    FITS file must have 'STACKED_FLUX' extension. This extension must have
    columns for 'lambda' and 'stacked_flux'. Wavelength array must be linearly
    and equally spaced. Uses linear interpolation.

    Parameters
    ----------
    fname: str
        Filename to read by ``fitsio``.
    comm: None or MPI.COMM_WORLD, default: None
    mpi_rank: int, default: 0

    Attributes
    ----------
    flux_interp: FastLinear1DInterp
        Flux interpolator.
    """

    def _read(self, fname):
        with fitsio.FITS(fname) as fts:
            data = fts['STACKED_FLUX'].read()

        waves = data['lambda']
        waves_0 = waves[0]
        dwave = waves[1] - waves[0]

        if not np.allclose(np.diff(waves), dwave):
            raise Exception(
                "Failed to construct flux calibration from "
                f"{fname}::wave is not equally spaced.")

        stacked_flux = np.array(data['stacked_flux'], dtype='d')
        stacked_flux[stacked_flux == 0] = 1

        return FastLinear1DInterp(waves_0, dwave, stacked_flux)

    def __init__(self, fname, comm=None, mpi_rank=0):
        self.stacked_flux_interp = mpi_fnc_bcast(
            self._read, comm, mpi_rank,
            f"Error loading FluxCalibrator from file {fname}.",
            fname)

    def apply(self, spectra_list):
        """ Apply the flux calibration by **only** scaling
        :attr:`forestflux <qsonic.spectrum.Spectrum.forestflux>` and
        :attr:`forestivar <qsonic.spectrum.Spectrum.forestivar>`. Smooth
        component must be set after this.

        Arguments
        ----------
        spectra_list: list(Spectrum)
            Spectrum objects to flux calibrate.
        """
        for spec in spectra_list:
            for arm, wave_arm in spec.forestwave.items():
                stacked_flux = self.stacked_flux_interp(wave_arm)
                spec.forestflux[arm] /= stacked_flux
                spec.forestivar[arm] *= stacked_flux**2



import logging
import numpy as np

from qsonic.mathtools import FastCubic1DInterp
from qsonic.continuum_models.base_continuum_model import BaseContinuumModel


class TrueContinuumModel(BaseContinuumModel):
    """True continuum model class for mock analysis. Uses fiducials for mean
    flux and varlss interpolation. Continuum is interpolated using a cubic
    spline.

    Parameters
    ----------
    meanflux_interp: FastLinear1DInterp
        Interpolator for mean flux. If fiducial is not set, this equals to 1.
    varlss_interp: FastLinear1DInterp or FastCubic1DInterp
        Cubic spline for var_lss if fitting. Linear if from file.
    """

    def __init__(self, meanflux_interp, varlss_interp):
        self.meanflux_interp = meanflux_interp
        self.varlss_interp = varlss_interp

    def fit_continuum(self, spec):
        """True continuum reduction. Uses fiducials for mean flux and varlss
        interpolation. Continuum is interpolated using a cubic spline.

        Arguments
        ---------
        spec: Spectrum
            Spectrum object to add true continuum.
        """
        w1 = spec.cont_params['true_data_w1']
        dwave = spec.cont_params['true_data_dwave']
        tcont = spec.cont_params['true_data']

        tcont_interp = FastCubic1DInterp(w1, dwave, tcont)

        for arm, wave_arm in spec.forestwave.items():
            cont_est = tcont_interp(wave_arm)
            cont_est *= self.meanflux_interp(wave_arm)
            spec.cont_params['cont'][arm] = cont_est

        spec.set_forest_weight(self.varlss_interp)
        spec.calc_continuum_chi2()

    def init_spectra(self, spectra_list):
        """ Initializes
        :attr:`cont_params <qsonic.spectrum.Spectrum.cont_params>` for a list
        of Spectrum objects.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects.
        """
        logging.info("Initializing true continuum.")

        for spec in spectra_list:
            spec.cont_params['method'] = 'true'
            spec.cont_params['valid'] = True
            spec.cont_params['x'] = np.zeros(1)
            spec.cont_params['xcov'] = np.eye(1)
            spec.cont_params['dof'] = spec.get_real_size()
            spec.cont_params['cont'] = {}

    def stacks_residual_flux(self):
        """:meth:`stack_spectra` stacks actual continuum values.

        Returns
        -------
        False: bool
        """
        return False



from .base_continuum_model import BaseContinuumModel
from .input_continuum_model import InputContinuumModel
from .picca_continuum_model import PiccaContinuumModel
from .true_continuum_model import TrueContinuumModel

__all__ = [
    "BaseContinuumModel", "PiccaContinuumModel", "InputContinuumModel",
    "TrueContinuumModel"
]



import argparse

import numpy as np

from qsonic.mathtools import _zero_function, _one_function, get_smooth_ivar


def add_wave_region_parser(parser=None):
    """ Adds wavelength analysis related arguments to parser. These
    arguments are grouped under 'Wavelength analysis region'. All of them
    come with defaults, none are required.

    Arguments
    ---------
    parser: argparse.ArgumentParser, default: None

    Returns
    ---------
    parser: argparse.ArgumentParser
    """
    if parser is None:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    wave_group = parser.add_argument_group('Wavelength analysis region')
    wave_group.add_argument(
        "--wave1", type=float, default=3600.,
        help="First observed wavelength edge.")
    wave_group.add_argument(
        "--wave2", type=float, default=6000.,
        help="Last observed wavelength edge.")
    wave_group.add_argument(
        "--forest-w1", type=float, default=1050.,
        help="First forest wavelength edge.")
    wave_group.add_argument(
        "--forest-w2", type=float, default=1180.,
        help="Last forest wavelength edge.")

    return parser


def generate_spectra_list_from_data(cat_by_survey, data):
    return [
        Spectrum.from_dictionary(catrow, data, idx)
        for idx, catrow in enumerate(cat_by_survey)
    ]


def valid_spectra(spectra_list):
    """Generator for continuum valid spectra."""
    return (spec for spec in spectra_list if spec.cont_params['valid'])


class Spectrum():
    """An object to represent one spectrum.

    Parameters
    ----------
    catrow: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog row.
    wave: dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        Dictionary of arrays specifying the wavelength grid. Static variable!
    flux: dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        Dictionary of arrays specifying the flux.
    ivar: dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        Dictionary of arrays specifying the inverse variance.
    mask: dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        Dictionary of arrays specifying the bitmask. Not stored
    reso: dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        Dictionary of 2D arrays specifying the resolution matrix.
    idx: int
        Index to access in flux, ivar, mask and reso that corresponds to the
        quasar in `catrow`.

    Attributes
    ----------
    rsnr: float
        Average SNR above Lya. Calculated in :meth:`set_forest_region`.
    mean_snr: dict(float)
        Mean signal-to-noise ratio in the forest.
    _f1, _f2: dict(int)
        Forest indices. Set up using :meth:`set_forest_region` method. Then use
        property functions to access forest wave, flux, ivar instead.
    cont_params: dict
        Continuum parameters. Initial estimates are constructed.
    """
    WAVE_LYA_A = 1215.67
    """float: Lya wavelength in A."""
    _wave = None
    """dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`): Common
    wavelength grid for **all** Spectra."""
    _coadd_wave = None
    """dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`): Common
    **coadded** wavelength grid for **all** Spectra."""
    _dwave = None
    """float: Wavelength spacing."""
    _blinding = None
    """str or None: Blinding. Must be set for certain data."""
    _fits_colnames = ['LAMBDA', 'DELTA', 'IVAR', 'WEIGHT', 'CONT']
    """list(str): Column names to save in delta files."""

    @staticmethod
    def _set_wave(wave, check_consistency=False):
        """Set the common wavelength grid.

        Arguments
        ---------
        check_consistency: bool
            Asserts each time key and values are the same if True.
        """
        if not Spectrum._wave:
            Spectrum._wave = wave.copy()
            arm = list(wave.keys())[0]
            wave_arm = wave[arm]
            Spectrum._dwave = wave_arm[1] - wave_arm[0]
        elif check_consistency:
            for arm, wave_arm in Spectrum._wave.items():
                assert (arm in wave.keys())
                assert (np.allclose(Spectrum._wave[arm], wave_arm))

    @staticmethod
    def _set_coadd_wave():
        if Spectrum._coadd_wave:
            return

        min_wave = np.min([wave[0] for wave in Spectrum._wave.values()])
        max_wave = np.max([wave[-1] for wave in Spectrum._wave.values()])

        # Z arm in mocks are shifted by 0.4 A, so we need these extra steps
        # to make sure dwave is preseved.
        nwaves = round((max_wave - min_wave) / Spectrum._dwave + 0.1)
        coadd_wave = np.linspace(
            min_wave, min_wave + nwaves * Spectrum._dwave, nwaves + 1)
        Spectrum._coadd_wave = {'brz': coadd_wave}

    @staticmethod
    def set_blinding(maxlastnight, args):
        """Set the blinding strategy.

        'LASTNIGHT' column, args.mock_analysis, args.forest_w1 decide the
        blinding strategy. Mock, side band and SV analyses are not blinded.

        Arguments
        ---------
        maxlastnight: int
            Maximum LASTNIGHT column of the entire quasar catalog.
        args: argparse.Namespace
            Should have ``mock_analysis (bool)`` and ``forest_w1 (floar)``.
        """
        # do not blind mocks or metal forests
        if args.mock_analysis or args.forest_w1 > Spectrum.WAVE_LYA_A:
            Spectrum._blinding = "none"
        # sv data, no blinding
        elif maxlastnight < 20210514:
            Spectrum._blinding = "none"
        elif maxlastnight < 20210801:
            Spectrum._blinding = "desi_m2"
        elif maxlastnight < 20220801:
            Spectrum._blinding = "desi_y1"
        else:
            Spectrum._blinding = "desi_y3"

        if Spectrum._blinding != "none":
            Spectrum._fits_colnames[1] = 'DELTA_BLIND'

        if not args.skip_resomat:
            Spectrum._fits_colnames.append('RESOMAT')

    @staticmethod
    def blinding_not_set():
        """bool: ``True`` if blinding is not set."""
        return Spectrum._blinding is None

    @classmethod
    def from_dictionary(cls, catrow, data, idx):
        """Create a Spectrum from dictionary. See :class:`Spectrum` for
        argument details.

        If ``cont`` key is present in ``data``, :attr:`cont_params` dictionary
        gains the following::

            cont_params['true_data_w1'] (float): First wavelength
            cont_params['true_data_dwave'] (float): Wavelength spacing
            cont_params['true_data'] (ndarray): True continuum

        Returns
        -------
        Spectrum
        """
        spec = cls(catrow, data['wave'], data['flux'], data['ivar'],
                   data['mask'], data['reso'], idx)

        if "cont" in data.keys():
            spec.cont_params['true_data_w1'] = data['cont']['w1']
            spec.cont_params['true_data_dwave'] = data['cont']['dwave']
            spec.cont_params['true_data'] = data['cont']['data'][idx]

        return spec

    def __init__(self, catrow, wave, flux, ivar, mask, reso, idx):
        self.catrow = catrow
        Spectrum._set_wave(wave)

        self._current_wave = Spectrum._wave
        self.flux = {}
        self.ivar = {}
        self.reso = {}

        self.rsnr = None
        self.mean_snr = None
        self._f1 = {}
        self._f2 = {}
        self._forestwave = {}
        self._forestflux = {}
        self._forestivar = {}
        self._forestivar_sm = {}
        self._forestreso = {}
        self._forestweight = {}

        self._smoothing_scale = 0

        for arm, wave_arm in self.wave.items():
            self.flux[arm] = flux[arm][idx].copy()
            self.ivar[arm] = ivar[arm][idx].copy()
            w = (mask[arm][idx] != 0) | np.isnan(self.flux[arm]) \
                | np.isnan(self.ivar[arm]) | (self.ivar[arm] < 0)
            self.flux[arm][w] = 0
            self.ivar[arm][w] = 0

            if not reso:
                continue
            elif reso[arm].ndim == 2:
                self.reso[arm] = reso[arm].copy()
            else:
                self.reso[arm] = reso[arm][idx].copy()

        self.cont_params = {}
        self.cont_params['method'] = ''
        self.cont_params['valid'] = False
        self.cont_params['x'] = np.array([1., 0.])
        self.cont_params['xcov'] = np.eye(2)
        self.cont_params['chi2'] = -1.
        self.cont_params['dof'] = 0
        self.cont_params['cont'] = {}

        self._set_rsnr()

    def _set_rsnr(self):
        """Calculates and sets SNR above Lya."""
        self.rsnr = 0
        rsnr_weight = 1e-6

        for arm, wave_arm in self.wave.items():
            # Calculate SNR above Lya
            ii1 = np.searchsorted(
                wave_arm, (1 + self.z_qso) * Spectrum.WAVE_LYA_A)
            weight = np.sqrt(self.ivar[arm][ii1:])
            self.rsnr += np.dot(self.flux[arm][ii1:], weight)
            rsnr_weight += np.sum(weight > 0)

        self.rsnr /= rsnr_weight

    def _set_forest_related_parameters(self):
        """Calculates the mean SNR in the forest region and an initial guess
        for the continuum amplitude."""
        self.cont_params['x'][0] = 0
        cont_params_weight = 1e-6

        self.mean_snr = {}

        for arm, ivar_arm in self.forestivar.items():
            flux_arm = self.forestflux[arm]
            w = flux_arm > 0

            self.cont_params['x'][0] += np.dot(flux_arm[w], ivar_arm[w])
            cont_params_weight += np.sum(ivar_arm[w])

            self.mean_snr[arm] = 0
            armpix = np.sum(ivar_arm > 0)
            if armpix == 0:
                continue

            self.mean_snr[arm] = np.dot(np.sqrt(ivar_arm), flux_arm) / armpix

        self.cont_params['x'][0] /= cont_params_weight

    def slice(self, arm, i1, i2):
        if i1 == 0 and i2 == self._forestwave[arm].size:
            return

        self._forestwave[arm] = self._forestwave[arm][i1:i2]
        self._forestflux[arm] = self._forestflux[arm][i1:i2]
        self._forestivar[arm] = self._forestivar[arm][i1:i2]
        self._forestivar_sm[arm] = self._forestivar_sm[arm][i1:i2]
        self._forestweight[arm] = self._forestweight[arm][i1:i2]

        if arm in self._forestreso:
            self._forestreso[arm] = self._forestreso[arm][:, i1:i2]

        if arm in self.cont_params['cont']:
            self.cont_params['cont'][arm] = \
                self.cont_params['cont'][arm][i1:i2]

    def set_forest_region(self, w1, w2, lya1, lya2):
        """ Sets slices for the forest region. Also calculates
        the mean SNR in the forest and an initial guess for the continuum
        amplitude. This is currently turned off, but it can mask outliers in
        each arm separately based on moving median statistics
        (see :func:`qsonic.mathtools.get_median_outlier_mask`).

        Arguments
        ---------
        w1, w2: float
            Observed wavelength range
        lya1, lya2: float
            Rest-frame wavelength for the forest
        """
        l1 = max(w1, (1 + self.z_qso) * lya1)
        l2 = min(w2, (1 + self.z_qso) * lya2)

        for arm, wave_arm in self.wave.items():
            # Slice to forest limits
            ii1, ii2 = np.searchsorted(wave_arm, [l1, l2])
            real_size_arm = np.sum(self.ivar[arm][ii1:ii2] > 0)
            if real_size_arm == 0:
                continue

            self._f1[arm], self._f2[arm] = ii1, ii2

            # See https://numpy.org/doc/stable/user/basics.copies.html
            # Slicing creates views, not copies. Bases of views are not removed
            # from memory!
            self._forestwave[arm] = wave_arm[ii1:ii2].copy()
            self._forestflux[arm] = self.flux[arm][ii1:ii2].copy()
            self._forestivar[arm] = self.ivar[arm][ii1:ii2].copy()
            if self.reso:
                self._forestreso[arm] = self.reso[arm][:, ii1:ii2].copy()

        self._forestivar_sm = self._forestivar
        self._forestweight = self._forestivar
        self._set_forest_related_parameters()

    def drop_arm(self, arm):
        self._forestwave.pop(arm, None)
        self._forestflux.pop(arm, None)
        self._forestivar.pop(arm, None)
        self._forestreso.pop(arm, None)
        self._forestivar_sm.pop(arm, None)
        self._forestweight.pop(arm, None)
        self.cont_params['cont'].pop(arm, None)

    def drop_short_arms(self, lya1=0, lya2=0, skip_ratio=0):
        """Arms that have less than ``skip_ratio`` pixels are removed from
        forest dictionary.

        Arguments
        ---------
        lya1, lya2: float
            Rest-frame wavelength for the forest
        skip_ratio: float
            Remove arms if they have less than this ratio of pixels
        """
        npixels_expected = (1 + self.z_qso) * (lya2 - lya1) / self.dwave
        npixels_expected = int(skip_ratio * npixels_expected) + 1
        short_arms = [arm for arm, ivar in self.forestivar.items()
                      if np.sum(ivar > 0) < npixels_expected]
        for arm in short_arms:
            self.drop_arm(arm)

    def remove_nonforest_pixels(self):
        """ Remove non-forest pixels from storage.

        This sets :attr:`flux`, :attr:`ivar` and :attr:`reso` to empty
        dictionary, but :attr:`wave` is not modified,
        since it is a static variable. Good practive is to loop using, e.g.,
        ``for arm, wave_arm in self.forestwave.items():``.
        """
        self._current_wave = {}
        self.flux = {}
        self.ivar = {}
        self.reso = {}

    def get_real_size(self):
        """
        Returns
        -------
        int: Sum of number of pixels with ``forestivar > 0`` for all arms.
        """
        size = 0
        for ivar_arm in self.forestivar.values():
            size += np.sum(ivar_arm > 0)

        return size

    def get_effective_meansnr(self):
        """ Calculate a weighted average of :attr:`mean_snr` over arms. Only
        call if :meth:`set_forest_region` has been called.

        .. math::
            \\langle\\mathrm{SNR}\\rangle = \\sum_{i} \\mathrm{SNR}_i^3 \\bigg/
            {\\sum_{i} \\mathrm{SNR}_i^2}

        Returns
        -------
        float: Effective mean SNR.
        """
        msnr_eff = 0
        norm = 1e-8

        for val in self.mean_snr.values():
            msnr_eff += val**3
            norm += val**2

        return msnr_eff / norm

    def is_long(self, dforest_wave, skip_ratio):
        """Determine if spectrum is long enough to be accepted.

        The condition is :meth:`get_real_size` > ``skip_ratio * npixels``,
        where ``npixels`` :math:`=(1 + z_\\mathrm{qso}) \\times`
        ``dforest_wave`` :math:`/ \\mathrm{d}\\lambda` and
        :math:`\\mathrm{d}\\lambda` is wavelength spacing in the observed frame
        in A.

        Arguments
        ---------
        dforest_wave: float
            Length of the forest in the rest-frame in A.
        skip_ratio: float
            Minimum ratio that needs to be present and unmasked to keep the
            spectrum.

        Returns
        -------
        bool
        """
        npixels = (1 + self.z_qso) * dforest_wave / self.dwave
        return self.get_real_size() > skip_ratio * npixels

    def set_smooth_forestivar(self, smoothing_size=16.):
        """ Set :attr:`forestivar_sm` to smoothed inverse variance. Before this
        call :attr:`forestivar_sm` points to :attr:`forestivar`. If
        ``smoothing_size <= 0``, smoothing is undone such that ivar_sm
        points to ivar. Also, :attr:`forestweight` points to this.

        ``smoothing_size`` is saved to a private :attr:`_smoothing_scale`
        variable for future use.

        Arguments
        ---------
        smoothing_size: float, default: 16
            Gaussian smoothing spread in A.
        """
        if smoothing_size <= 0:
            self._smoothing_scale = 0
            self._forestivar_sm = self._forestivar
        else:
            self._smoothing_scale = smoothing_size
            sigma_pix = smoothing_size / self.dwave
            self._forestivar_sm = {
                arm: get_smooth_ivar(ivar_arm, sigma_pix)
                for arm, ivar_arm in self.forestivar.items()
            }

        self._forestweight = self._forestivar_sm

    def set_forest_weight(
            self, varlss_interp=_zero_function, eta_interp=_one_function
    ):
        """ Sets :attr:`forestweight` for a given var_lss and eta correction.
        Always uses :attr:`forestivar_sm`, which is not actually smoothed if
        :meth:`set_smooth_forestivar` is not called.

        .. math::

            w = i / (\\eta + i \\sigma^2_\\mathrm{LSS} C^2),

        where i is IVAR and C is the continuum.

        Arguments
        ---------
        varlss_interp: Callable[[ndarray], ndarray], default: 0
            LSS variance interpolator.
        eta_interp: Callable[[ndarray], ndarray], default: 1
            eta interpolator.
        """
        if not self.cont_params['valid'] or not self.cont_params['cont']:
            self._forestweight = self._forestivar_sm
            return

        self._forestweight = {}
        for arm, wave_arm in self.forestwave.items():
            cont_est = self.cont_params['cont'][arm]
            var_lss = varlss_interp(wave_arm) * cont_est**2
            eta = eta_interp(wave_arm)
            ivar_arm = self.forestivar_sm[arm]
            self._forestweight[arm] = ivar_arm / (eta + ivar_arm * var_lss)

    def calc_continuum_chi2(self):
        """ Calculate the chi2 of the continuum fitting. This is just a sum
        of weight * (flux - cont)^2.

        Returns
        -------
        chi2: float
        """
        if not self.cont_params['valid'] or not self.cont_params['cont']:
            self.cont_params['chi2'] = -1
            return -1

        chi2 = 0
        for arm, wave_arm in self.forestwave.items():
            cont_est = self.cont_params['cont'][arm]
            weight = self.forestweight[arm]
            flux = self.forestflux[arm]
            chi2 += np.dot(weight, (flux - cont_est)**2)

        self.cont_params['chi2'] = chi2

        return chi2

    def simple_coadd(self):
        """Coadding without continuum and var_lss terms on the full spectrum.
        Weights, forests etc. will not be set. Replaces :attr:`wave`,
        :attr:`flux`, :attr:`ivar` and :attr:`reso` attributes with
        dictionaries that has a single arm ``brz`` as key to access the coadded
        data.

        We first set the static coadded wavelength grid if it is not set.
        Then we add each arm using :attr:`ivar` (not smoothed). If :attr:`reso`
        is set, we coadd the resolution matrix using the same inverse variance
        weights. If these weights are zero for both arms, resolution matrix
        coadding reverts to equal weights. Final private assignments are done
        at the end to keep using arm by arm values of :attr:`ivar`.
        """
        Spectrum._set_coadd_wave()

        min_wave = Spectrum._coadd_wave['brz'][0]
        nwaves = Spectrum._coadd_wave['brz'].size

        coadd_flux = np.zeros(nwaves)
        coadd_ivar = np.zeros(nwaves)
        coadd_norm = np.zeros(nwaves)

        idxes = {}
        for arm, wave_arm in self.wave.items():
            idx = ((wave_arm - min_wave) / self.dwave + 0.1).astype(int)
            idxes[arm] = idx

            weight = self.ivar[arm]

            var = np.zeros_like(weight)
            w = self.ivar[arm] > 0
            var[w] = 1 / self.ivar[arm][w]

            coadd_flux[idx] += weight * self.flux[arm]
            coadd_ivar[idx] += weight**2 * var
            coadd_norm[idx] += weight

        w = coadd_norm > 0
        coadd_flux[w] /= coadd_norm[w]
        coadd_ivar[w] = coadd_norm[w]**2 / coadd_ivar[w]

        if self.reso:
            max_ndia = np.max([reso.shape[0] for reso in self.reso.values()])
            coadd_reso = np.zeros((max_ndia, nwaves))
            coadd_norm *= 0

            for arm, reso_arm in self.reso.items():
                weight = self.ivar[arm].copy()
                weight[weight == 0] = 1e-8

                ddia = max_ndia - reso_arm.shape[0]
                # Assumption ddia cannot be odd
                ddia = ddia // 2
                if ddia > 0:
                    reso_arm = np.pad(reso_arm, ((ddia, ddia), (0, 0)))

                coadd_reso[:, idxes[arm]] += weight * reso_arm
                coadd_norm[idxes[arm]] += weight

            coadd_reso /= coadd_norm
            self.reso = {'brz': coadd_reso}

        self._current_wave = Spectrum._coadd_wave
        self.flux = {'brz': coadd_flux}
        self.ivar = {'brz': coadd_ivar}

    def coadd_arms_forest(
            self, varlss_interp=_zero_function, eta_interp=_one_function
    ):
        """ Coadds different arms using :attr:`forestweight`. Interpolators are
        needed to reset :attr:`forestweight`.

        Replaces ``forest`` variables and ``cont_params['cont']`` with a
        dictionary that has a single arm ``brz`` as key to access coadded data.

        Arguments
        ---------
        varlss_interp: Callable[[ndarray], ndarray], default: 0
            LSS variance interpolator or function.
        eta_interp: Callable[[ndarray], ndarray], default: 1
            eta interpolator or function.
        """
        min_wave = np.min([wave[0] for wave in self.forestwave.values()])
        max_wave = np.max([wave[-1] for wave in self.forestwave.values()])

        # nwaves = round((max_wave - min_wave) / self.dwave) + 1
        # coadd_wave = np.linspace(
        #     min_wave, min_wave + (nwaves - 1) * self.dwave, nwaves)
        nwaves = int((max_wave - min_wave) / self.dwave + 0.1) + 1
        coadd_wave = np.linspace(min_wave, max_wave, nwaves)
        coadd_flux = np.zeros(nwaves)
        coadd_ivar = np.zeros(nwaves)
        coadd_norm = np.zeros(nwaves)

        idxes = {}
        for arm, wave_arm in self.forestwave.items():
            idx = ((wave_arm - min_wave) / self.dwave + 0.1).astype(int)
            idxes[arm] = idx

            weight = self.forestweight[arm]

            var = np.zeros_like(weight)
            w = self.forestivar[arm] > 0
            var[w] = 1 / self.forestivar[arm][w]

            coadd_flux[idx] += weight * self.forestflux[arm]
            coadd_ivar[idx] += weight**2 * var
            coadd_norm[idx] += weight

        w = coadd_norm > 0
        coadd_flux[w] /= coadd_norm[w]
        coadd_ivar[w] = coadd_norm[w]**2 / coadd_ivar[w]

        self._forestwave = {'brz': coadd_wave}
        self._forestflux = {'brz': coadd_flux}
        self._forestivar = {'brz': coadd_ivar}

        if self.cont_params['cont']:
            coadd_cont = np.empty(nwaves)

            for arm, idx in idxes.items():
                # continuum needs not weighting
                coadd_cont[idx] = self.cont_params['cont'][arm]

            self.cont_params['cont'] = {'brz': coadd_cont}

        if self.forestreso:
            max_ndia = np.max(
                [reso.shape[0] for reso in self.forestreso.values()])
            coadd_reso = np.zeros((max_ndia, nwaves))
            coadd_norm.fill(0)

            for arm, reso_arm in self.forestreso.items():
                weight = self.forestweight[arm].copy()
                weight[weight == 0] = 1e-8

                ddia = max_ndia - reso_arm.shape[0]
                # Assumption ddia cannot be odd
                ddia = ddia // 2
                if ddia > 0:
                    reso_arm = np.pad(reso_arm, ((ddia, ddia), (0, 0)))

                coadd_reso[:, idxes[arm]] += weight * reso_arm
                coadd_norm[idxes[arm]] += weight

            coadd_reso /= coadd_norm
            self._forestreso = {'brz': coadd_reso}

        self.set_smooth_forestivar(self._smoothing_scale)
        self.set_forest_weight(varlss_interp, eta_interp)

        mean_snr = np.dot(
            np.sqrt(coadd_ivar), coadd_flux) / np.sum(coadd_ivar > 0)
        self.mean_snr = {'brz': mean_snr}

    def mean_resolution(self, arm, weight=None):
        """ Returns the weighted mean Gaussian sigma of the spectrograph
        resolution in the forest.

        Arguments
        ---------
        arm: str
            Arm.
        weight: None or ndarray, default: None
            Weights. If ``None``, :attr:`forestweight` is used.

        Returns
        -------
        mean_reso: float or None
            Gaussian sigma. None if forestreso is not set.
        """
        if not self.forestreso:
            return None

        if weight is None:
            weight = self.forestweight[arm]

        total_weight = np.sum(weight)
        reso = np.dot(self.forestreso[arm], weight) / total_weight
        lambda_eff = np.dot(self.forestwave[arm], weight) / total_weight

        central_idx = reso.argmax()
        off_idx = np.array([-2, -1, 1, 2], dtype=int)
        ratios = reso[central_idx] / reso[central_idx + off_idx]
        ratios = np.log(ratios)
        w2 = ratios > 0
        norm = np.sum(w2)
        new_ratios = np.zeros_like(ratios)
        new_ratios[w2] = 1. / np.sqrt(ratios[w2])

        rms_in_pixel = np.abs(off_idx).dot(new_ratios) / np.sqrt(2.) / norm
        return rms_in_pixel * 3e5 * self.dwave / lambda_eff

    def write(self, fts_file):
        """Writes each arm to FITS file separately.

        Writes 'LAMBDA', 'DELTA', 'IVAR', 'WEIGHT', 'CONT' columns and
        'RESOMAT' column if resolution matrix is present to extension name
        ``targetid-arm``. FITS file must be initialized before.
        Each arm has its own `MEANSNR`.

        Arguments
        ---------
        fts_file: FITS file
            The file handler, not filename.
        """
        hdr_dict = {
            'LOS_ID': self.targetid,
            'TARGETID': self.targetid,
            'RA': np.radians(self.ra),
            'DEC': np.radians(self.dec),
            'Z': self.z_qso,
            'BLINDING': Spectrum._blinding,
            'WAVE_SOLUTION': "lin",
            'MEANSNR': 0.,
            'RSNR': self.rsnr,
            'DELTA_LAMBDA': self.dwave,
            'SMSCALE': self._smoothing_scale
        }

        for key in set(
                ['TILEID', 'FIBER', 'PETAL_LOC', 'EXPID', 'NIGHT']
        ).intersection(self.catrow.dtype.names):
            hdr_dict[key] = self.catrow[key]

        if 'EXPID' in self.catrow.dtype.names:
            expid = f"-{self.catrow['EXPID']}"
        else:
            expid = ""

        for arm, wave_arm in self.forestwave.items():
            if self.mean_snr[arm] == 0:
                continue

            hdr_dict['MEANSNR'] = self.mean_snr[arm]

            cont_est = self.cont_params['cont'][arm]
            delta = self.forestflux[arm] / cont_est - 1
            ivar = self.forestivar[arm] * cont_est**2
            weight = self.forestweight[arm] * cont_est**2
            delta[ivar == 0] = 0

            cols = [wave_arm, delta, ivar, weight, cont_est]
            if self.forestreso:
                hdr_dict['MEANRESO'] = self.mean_resolution(arm)
                cols.append(self.forestreso[arm].T.astype('f8'))

            fts_file.write(
                cols, names=Spectrum._fits_colnames, header=hdr_dict,
                extname=f"{self.targetid}-{arm}{expid}")

    @property
    def z_qso(self):
        """float: Quasar redshift."""
        return self.catrow['Z']

    @property
    def targetid(self):
        """int: Unique TARGETID identifier."""
        return self.catrow['TARGETID']

    @property
    def hpix(self):
        """int: Healpix."""
        return self.catrow['HPXPIXEL']

    @property
    def ra(self):
        """float: Right ascension."""
        return self.catrow['RA']

    @property
    def dec(self):
        """float: Declination"""
        return self.catrow['DEC']

    @property
    def wave(self):
        """dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`): Original
        wavelength grid in A."""
        return self._current_wave

    @property
    def dwave(self):
        """float: Wavelength step size in A."""
        return Spectrum._dwave

    @property
    def forestwave(self):
        """dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`): Forest
        wavelength field in A."""
        return self._forestwave

    @property
    def forestflux(self):
        """dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`): Forest
        flux field."""
        return self._forestflux

    @property
    def forestivar(self):
        """dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`): Forest
        inverse variance field."""
        return self._forestivar

    @property
    def forestivar_sm(self):
        """dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`): Forest
        smoothed inverse variance field.

        Initially equal to :attr:`.forestivar`. Smoothed if
        :meth:`.set_smooth_forestivar` is called."""
        return self._forestivar_sm

    @property
    def forestweight(self):
        """dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`): Forest
        weight field. Initially equal to :attr:`.forestivar`."""
        return self._forestweight

    @property
    def forestreso(self):
        """dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`): Resolution
        matrix in the forest."""
        return self._forestreso


class Delta():
    """An object to read one delta from HDU.

    Parameters
    ----------
    hdu: fitsio.TableHDU
        Table containing delta data.

    Raises
    ---------
    RuntimeError
        If ``hdu`` doesn't have neither "LAMBDA" nor "LOGLAM" columns.
    RuntimeError
        If ``hdu`` doesn't have neither "DELTA" nor "DELTA_BLIND" columns.
    RuntimeError
        If ``hdu`` doesn't have none of "MOCKID", "TARGETID" and "THING_ID"
        header keys.

    Attributes
    ----------
    wave: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Wavelength array in A.
    delta: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Deltas.
    ivar: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Inverse variance.
    weight: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Weights, which includes var_lss.
    cont: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Continuum.
    reso: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Resolution matrix. ``None`` if not present.
    header: FITS header
        Header.
    targetid: int
        TARGETID, MOCKID or THING_ID from header.
    mean_snr: float
        MEANSNR from header.
    """
    _accepted_wave_columns = set(["LAMBDA", "LOGLAM"])
    """set: Supported column names for wavelength."""
    _accepted_delta_columns = set(['DELTA', 'DELTA_BLIND'])
    """set: Supported column names for delta."""
    _accepted_targetid_keys = set(['MOCKID', 'TARGETID', 'THING_ID'])
    """set: Supported header keys for unique ID."""
    _accepted_colums_map = {
        "wave": _accepted_wave_columns,
        "delta": _accepted_delta_columns
    }

    @staticmethod
    def _check_hdu(colnames, attr):
        req_map = Delta._accepted_colums_map[attr]
        key = req_map.intersection(colnames)
        if not key:
            raise RuntimeError(
                "One of these must be present in delta files: "
                f"{', '.join(req_map)} for {attr}!")

        return key.pop()

    def __init__(self, hdu):
        self.header = hdu.read_header()
        key = Delta._accepted_targetid_keys.intersection(self.header.keys())
        if not key:
            raise RuntimeError(
                "One of these must be present in delta file header: "
                f"{', '.join(Delta._accepted_targetid_keys)} for TARGETID!")

        key = key.pop()
        self.targetid = self.header[key]
        self.mean_snr = self.header['MEANSNR']

        colnames = hdu.get_colnames()
        data = hdu.read()

        key = Delta._check_hdu(colnames, "wave")
        if key == "LOGLAM":
            self.wave = 10**data['LOGLAM']
        else:
            self.wave = data[key].astype("f8")

        key = Delta._check_hdu(colnames, "delta")
        self._is_blinded = key == "DELTA_BLIND"
        self.delta = data[key].astype("f8")
        self.ivar = data['IVAR'].astype("f8")
        self.weight = data['WEIGHT'].astype("f8")
        self.cont = data['CONT'].astype("f8")

        self.delta[self.ivar == 0] = 0

        if 'RESOMAT' in colnames:
            self.reso = data['RESOMAT'].T.astype("f8")
        else:
            self.reso = None

    def write(self, fts_file):
        """Writes to FITS file. This function is aimed at saving coadded
        deltas.

        Writes 'LAMBDA', 'DELTA', 'IVAR', 'WEIGHT', 'CONT' columns and
        'RESOMAT' column if resolution matrix is present to extension name
        ``targetid``. FITS file must be initialized before. Note that ``arm``
        is lost in the extension name.

        Arguments
        ---------
        fts_file: FITS file
            The file handler, not filename.
        """
        hdr_dict = self.header

        cols = [self.wave, self.delta, self.ivar, self.weight, self.cont]
        names = ['LAMBDA', 'DELTA', 'IVAR', 'WEIGHT', 'CONT']

        if self._is_blinded:
            names[1] = 'DELTA_BLIND'

        if self.reso is not None:
            cols.append(self.reso.T)
            names.append('RESOMAT')

        fts_file.write(
            cols, names=names, header=hdr_dict,
            extname=f"{self.targetid}")

    def _coadd_reso(self, other, nwaves, idxes):
        max_ndia = max(self.reso.shape[0], other.reso.shape[0])
        coadd_reso = np.zeros((max_ndia, nwaves))
        coadd_norm = np.zeros(nwaves)

        for j, obj in enumerate([self, other]):
            weight = obj.weight.copy()
            weight[weight == 0] = 1e-8

            ddia = max_ndia - obj.reso.shape[0]
            # Assumption ddia cannot be odd
            ddia = ddia // 2
            if ddia > 0:
                obj.reso = np.pad(obj.reso, ((ddia, ddia), (0, 0)))

            coadd_reso[:, idxes[j]] += weight * obj.reso
            coadd_norm[idxes[j]] += weight

        coadd_reso /= coadd_norm
        self.reso = coadd_reso

    def coadd(self, other, dwave=0.8):
        min_wave = min(self.wave[0], other.wave[0])
        max_wave = max(self.wave[-1], other.wave[-1])

        nwaves = int((max_wave - min_wave) / dwave + 0.1) + 1
        coadd_wave = np.linspace(min_wave, max_wave, nwaves)
        coadd_delta = np.zeros(nwaves)
        coadd_ivar = np.zeros(nwaves)
        coadd_lss = np.zeros(nwaves)
        coadd_cont = np.zeros(nwaves)
        coadd_norm = np.zeros(nwaves)

        idxes = [None, None]
        for j, obj in enumerate([self, other]):
            i0 = ((obj.wave[0] - min_wave) / dwave + 0.1).astype(int)
            idx = np.s_[i0:i0 + obj.wave.size]
            idxes[j] = idx

            var = np.zeros_like(obj.weight)
            w = (obj.weight > 0) & (obj.ivar > 0)
            var[w] = 1 / obj.ivar[w]

            coadd_delta[idx] += obj.weight * obj.delta
            coadd_cont[idx] += obj.weight * obj.cont
            coadd_ivar[idx] += obj.weight**2 * var
            coadd_lss[idx] += 1 - obj.weight * var
            coadd_norm[idx] += obj.weight

        w = coadd_norm > 0
        coadd_delta[w] /= coadd_norm[w]
        coadd_cont[w] /= coadd_norm[w]
        coadd_lss[w] /= coadd_norm[w]
        coadd_ivar[w] = coadd_norm[w]**2 / coadd_ivar[w]

        if self.reso is not None:
            self._coadd_reso(other, nwaves, idxes)

        self.wave = coadd_wave
        self.delta = coadd_delta
        self.ivar = coadd_ivar
        self.weight = self.ivar / (1 + self.ivar * coadd_lss)
        self.cont = coadd_cont

        self.mean_snr = np.dot(
            np.sqrt(coadd_ivar), coadd_delta + 1) / np.sum(coadd_ivar > 0)
        self.header['MEANSNR'] = self.mean_snr

    @property
    def ra(self):
        """float: Right ascension in degrees."""
        return np.degrees(self.header['RA'])

    @property
    def dec(self):
        """float: Declination in degrees"""
        return np.degrees(self.header['DEC'])



import argparse
import pytest
from unittest import TestCase

import numpy as np
import numpy.testing as npt

from qsonic import QsonicException
import qsonic.mpi_utils


class TestMPIUtils(TestCase):
    @pytest.mark.mpi(min_size=2)
    def test_mpi_parse(self):
        from mpi4py import MPI
        from qsonic.io import add_io_parser
        comm = MPI.COMM_WORLD
        assert comm.size > 0

        mpi_rank = comm.Get_rank()
        parser = argparse.ArgumentParser()
        add_io_parser(parser)

        options = "--input-dir indir --catalog incat -o outdir".split(' ')
        args = qsonic.mpi_utils.mpi_parse(parser, comm, mpi_rank, options)
        assert (args.input_dir == "indir")
        assert (args.catalog == "incat")
        assert (args.outdir == "outdir")

        with pytest.raises(SystemExit):
            options = "--catalog incat -o outdir".split(' ')
            qsonic.mpi_utils.mpi_parse(parser, comm, mpi_rank, options)

        # Test logic functions
        from qsonic.scripts.qsonic_fit import (
            get_parser, args_logic_fnc_qsonic_fit)

        parser = get_parser()

        options = ("--input-dir indir --catalog incat -o outdir "
                   "--mock-analysis --continuum-model true "
                   "--fiducial-meanflux mflux "
                   "--fiducial-varlss varlss").split(' ')
        args = qsonic.mpi_utils.mpi_parse(parser, comm, mpi_rank, options,
                                          args_logic_fnc_qsonic_fit)
        assert args.continuum_model == "true"
        assert args.true_continuum
        assert args.mock_analysis

        with pytest.raises(SystemExit):
            options = ("--input-dir indir --catalog incat -o outdir "
                       "--continuum-model true").split(' ')
            qsonic.mpi_utils.mpi_parse(parser, comm, mpi_rank, options,
                                       args_logic_fnc_qsonic_fit)

        with pytest.raises(SystemExit):
            qsonic.mpi_utils.mpi_parse(parser, comm, mpi_rank, ['-h'],
                                       args_logic_fnc_qsonic_fit)

    def test_mpi_fnc_bcast(self):
        matrix = np.arange(20).reshape(4, 5)
        u, s, vh = np.linalg.svd(matrix)
        result = qsonic.mpi_utils.mpi_fnc_bcast(
            np.linalg.svd, None, 0, "Error", matrix)

        npt.assert_allclose(result[0], u)
        npt.assert_allclose(result[1], s)
        npt.assert_allclose(result[2], vh)

        s = np.linalg.svd(matrix, compute_uv=False)
        result = qsonic.mpi_utils.mpi_fnc_bcast(
            np.linalg.svd, None, 0, "Error", matrix, compute_uv=False)

        npt.assert_allclose(result, s)

        with pytest.raises(QsonicException):
            qsonic.mpi_utils.mpi_fnc_bcast(np.zeros, None, 0, "Error", -5)

    def test_balance_load(self):
        split_catalog = [
            np.ones(3), 2 * np.ones(4), 3 * np.ones(5), 4 * np.ones(1)]

        sorted_catalog = [
            3 * np.ones(5), 2 * np.ones(4), np.ones(3), 4 * np.ones(1)]

        mpi_size = 3
        q0, q1, q2 = qsonic.mpi_utils.balance_load(split_catalog, mpi_size)

        for idx in range(len(split_catalog)):
            npt.assert_allclose(split_catalog[idx], sorted_catalog[idx])
        assert (len(q0) == 1)
        assert (len(q1) == 1)
        assert (len(q2) == 2)
        npt.assert_allclose(q0, 3)
        npt.assert_allclose(q1, 2)


if __name__ == '__main__':
    pytest.main()



import pytest

import fitsio
import numpy as np
import numpy.testing as npt

import qsonic.spectrum
import qsonic.calibration


@pytest.fixture()
def my_setup_attributes(tmp_path):
    fname = tmp_path / "attributes.fits"

    with fitsio.FITS(fname, 'rw', clobber=True) as fts:
        wavelength = np.linspace(3600, 9800, 775)
        stacked_flux = 5. * np.ones(wavelength.size)
        fts.write(
            [wavelength, stacked_flux],
            names=['lambda', 'stacked_flux'],
            extname='STACKED_FLUX')

        wavelength = np.linspace(3600, 9800, 50)
        ones_array = np.ones(wavelength.size)
        fts.write(
            [wavelength, ones_array, ones_array, ones_array * 2, ones_array],
            names=['lambda', 'var_lss', 'e_var_lss', 'eta', 'e_eta'],
            extname='VAR_FUNC')

    yield fname


def test_calibrations(setup_data, my_setup_attributes):
    fname = my_setup_attributes
    ncal = qsonic.calibration.NoiseCalibrator(fname)
    ncal_plus = qsonic.calibration.NoiseCalibrator(fname, add_varlss=True)
    fcal = qsonic.calibration.FluxCalibrator(fname)

    cat_by_survey, _, data = setup_data(1)
    spectra_list = qsonic.spectrum.generate_spectra_list_from_data(
        cat_by_survey, data)
    spec = spectra_list[0]
    spec.set_forest_region(3600., 6000., 1050., 1300.)
    f0 = spec.flux['B'][0]

    ncal.apply(spectra_list)
    npt.assert_allclose(spec.forestivar['B'], 0.5)
    npt.assert_allclose(spec.forestivar['R'], 0.5)

    with pytest.raises(Exception):
        ncal_plus.apply(spectra_list)

    spec.cont_params['cont'] = {
        arm: np.ones_like(flux) for arm, flux in spec.forestflux.items()
    }
    ncal_plus.apply(spectra_list)
    npt.assert_allclose(spec.forestivar['B'], 0.2)
    npt.assert_allclose(spec.forestivar['R'], 0.2)

    fcal.apply(spectra_list)
    npt.assert_allclose(spec.forestflux['B'], f0 / 5)
    npt.assert_allclose(spec.forestflux['R'], f0 / 5)



import argparse
import functools
import glob
import logging

from collections import defaultdict
from multiprocessing import Pool

import fitsio
import healpy

import qsonic.io


def get_parser():
    """Constructs the parser needed for the script.

    Returns
    -------
    parser: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        "--input-dirs", "-i", nargs='+', type=str, required=True,
        help=('Directories that have delta-*.fits files. Multiple arguments '
              'are passed with a space')
    )
    parser.add_argument(
        "--output-dir", "-o", type=str,
        help="Output directory.", required=True)
    parser.add_argument(
        "--nside", type=int, default=8, help='Output healpix nside')
    parser.add_argument(
        "--nest", action="store_true", help="Use NESTED pixelization.")
    parser.add_argument(
        "--nproc", type=int, default=None, help="Number of processes.")
    return parser


def read_dirs_to_dict(directories, nproc):
    """Reads all directories full of ``delta-*.fits`` files, and coadds them
    into ``forest_dict`` dictionary.

    Arguments
    ---------
    directories: list(str)
        List of directories to search for `delta-*.fits`` files.
    nproc: int
        Number of processes.

    Returns
    -------
    forest_dict: dict(Delta)
        Dictionary of Delta objects by ``TARGETID`` as key.
    """
    logging.info("Making list of all delta files.")
    fnames = []
    for directory in directories:
        fnames.extend(glob.glob(f"{directory}/delta-*.fits"))

    if not fnames:
        raise Exception("No delta files are found!")

    logging.info(f"There are {len(fnames)} files.")
    with Pool(processes=nproc) as pool:
        deltas = pool.map(qsonic.io.read_deltas, fnames)

    logging.info("Coadding deltas.")
    forest_dict = {}
    for onelist in deltas:
        for delta in onelist:
            if delta.targetid in forest_dict:
                forest_dict[delta.targetid].coadd(delta)
            else:
                forest_dict[delta.targetid] = delta

    return forest_dict


def write_one_hpx(hpx, list_deltas, output_dir):
    """Writes to one healpix delta file.

    Arguments
    ---------
    hpx: int
        Healpixel.
    list_deltas: list(Delta)
        List of delta files.
    output_dir: str
        Output directory.
    """
    with fitsio.FITS(
            f"{output_dir}/delta-{hpx}.fits", 'rw', clobber=True
    ) as fts:
        for delta in list_deltas:
            delta.write(fts)


def main():
    args = get_parser().parse_args()

    logging.basicConfig(
        format='%(asctime)s - %(levelname)s: %(message)s',
        datefmt='%Y/%m/%d %I:%M:%S %p',
        level=logging.DEBUG)

    forest_dict = read_dirs_to_dict(args.input_dirs, args.nproc)

    logging.info("Grouping into healpixels.")
    forest_by_hpx = defaultdict(list)
    for targetid, delta in forest_dict.items():
        hpx = healpy.ang2pix(
            args.nside, delta.ra, delta.dec, lonlat=True, nest=args.nest)
        forest_by_hpx[hpx].append(delta)

    logging.info(
        f"There are {len(forest_by_hpx)} healpixels. Writing to files.")

    _fnc = functools.partial(write_one_hpx, output_dir=args.output_dir)

    with Pool(processes=args.nproc) as pool:
        pool.starmap(_fnc, forest_by_hpx.items())



"""Mathematical utility objects and functions."""
import numpy as np
from numba import njit
from scipy.ndimage import median_filter

from qsonic import QsonicException


def _zero_function(x):
    return 0


def _one_function(x):
    return 1


@njit("f8[:](f8[:], f8, f8, f8[:])")
def _fast_eval_interp1d_lin(x, xp0, dxp, fp):
    """JIT fast linear interpolation."""
    xx = (x - xp0) / dxp
    idx = np.clip(xx, 0, fp.size - 1 - 1e-8).astype(np.int_)

    d_idx = xx - idx
    y1, y2 = fp[idx], fp[idx + 1]

    return y1 * (1 - d_idx) + y2 * d_idx


@njit("f8[:](f8[:], f8, f8, f8[:], f8[:])")
def _fast_eval_interp1d_cubic(x, xp0, dxp, fp, y2p):
    """JIT fast cubic spline."""
    xx = (x - xp0) / dxp
    idx = np.clip(xx, 0, fp.size - 1 - 1e-8).astype(np.int_)

    d_idx = xx - idx
    a, b = 1 - d_idx, d_idx
    y1, y2 = fp[idx], fp[idx + 1]
    ypp1, ypp2 = y2p[idx], y2p[idx + 1]

    r1 = a * y1 + b * y2
    r2 = ((a + 1) * ypp1 + (b + 1) * ypp2) * (-a * b * dxp**2 / 6.)

    return r1 + r2


@njit("f8[:](f8[:], f8)")
def _spline_cubic_natural(fp, dxp):
    """ Constructs the second derivative array with natural boundary
    conditions.

    As coded in Numerical Recipes in C Chaper 3.3, but specialized for equally
    spaced input data.

    Arguments
    ---------
    fp: ndarray
        1D data points array to interpolate.
    dxp: float
        Spacing of x points.

    Returns
    -------
    y2p: ndarray
        Second derivative array. Same size as ``fp``.
    """
    y2p = np.empty(fp.size)
    u = np.empty(fp.size - 1)
    u[0] = 0
    y2p[0] = y2p[-1] = 0

    sig = 0.5

    for ii in range(1, fp.size - 1):
        p = sig * y2p[ii - 1] + 2.
        y2p[ii] = (sig - 1) / p
        u[ii] = (fp[ii + 1] + fp[ii - 1] - 2 * fp[ii]) / dxp
        u[ii] = (3 * u[ii] / dxp - sig * u[ii - 1]) / p

    for ii in range(fp.size - 2, 0, -1):
        y2p[ii] = y2p[ii] * y2p[ii + 1] + u[ii]

    return y2p


@njit("f8[:](f8[:], f8)")
def _spline_cubic_notaknot(fp, dxp):
    """Constructs the second derivative array with not-a-knot boundary
    conditions:

    .. math::

        y''_0 - 2 y''_1 + y''_2 = 0

    Arguments
    ---------
    fp: ndarray
        1D data points array to interpolate.
    dxp: float
        Spacing of x points.

    Returns
    -------
    y2p: ndarray
        Second derivative array. Same size as ``fp``.
    """
    N = fp.size
    y2p = np.empty(N)
    # Solve the submatrix 1:-1 using precalculated u and p values
    u = np.array([
        6.000000000000000000e+00, 4.000000000000000000e+00,
        3.750000000000000000e+00, 3.733333333333333393e+00,
        3.732142857142857206e+00, 3.732057416267942518e+00,
        3.732051282051282115e+00, 3.732050841635176752e+00,
        3.732050810014727382e+00, 3.732050807744481613e+00,
        3.732050807581485330e+00, 3.732050807569782691e+00,
        3.732050807568942474e+00, 3.732050807568882078e+00,
        3.732050807568877637e+00, 3.732050807568877193e+00])
    p = np.array([
        1.666666666666666574e-01, 2.500000000000000000e-01,
        2.666666666666666630e-01, 2.678571428571428492e-01,
        2.679425837320574266e-01, 2.679487179487179405e-01,
        2.679491583648230812e-01, 2.679491899852724512e-01,
        2.679491922555185535e-01, 2.679491924185148921e-01,
        2.679491924302174755e-01, 2.679491924310576922e-01,
        2.679491924311180329e-01, 2.679491924311223627e-01,
        2.679491924311226958e-01])

    for i in range(1, N - 1):
        y2p[i] = fp[i - 1] - 2 * fp[i] + fp[i + 1]

    for i in range(1, min(N - 3, p.size + 1)):
        y2p[i + 1] -= p[i - 1] * y2p[i]
    for i in range(p.size + 1, N - 3):
        y2p[i + 1] -= p[-1] * y2p[i]

    y2p[-2] /= u[0]
    for i in range(N - 4, u.size - 1, -1):
        y2p[i + 1] = (y2p[i + 1] - y2p[i + 2]) / u[-1]
    for i in range(min(N - 4, u.size - 1), 0, -1):
        y2p[i + 1] = (y2p[i + 1] - y2p[i + 2]) / u[i]

    y2p[1] /= u[0]
    y2p[0] = 2 * y2p[1] - y2p[2]
    y2p[-1] = 2 * y2p[-2] - y2p[-3]
    y2p *= 6.0 / dxp**2

    return y2p


@njit("f8[:](f8[:], f8[:])")
def mypoly1d(coef, x):
    """ My simple power series polynomial calculator.

    Arguments
    ---------
    coef: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Coefficient array in increasing power starting with the constant.
    x: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Array to calculate polynomial.

    Returns
    ---------
    results: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Polynomial calculated at x.
    """
    results = np.zeros_like(x)
    for i, a in enumerate(coef):
        results += a * x**i
    return results


@njit("f8[:, :, :](f8[:], f8[:], f8[:, :, :])")
def block_covariance_of_square(mean, var, cov):
    """ Return the block covariance of x^2, i.e.
    :math:`<x_i^2 x_j^2> - <x_i^2><x_j^2>`. Compatible with ``blockdim``
    argument of :meth:`SubsampleCov.get_mean_n_cov`.

    .. math::

        Cov[x_i^2, x_j^2] &= Var[x_i] Var[x_j] + Cov(x_i, x_j)^2 \\\\
        &+ 2 Cov(x_i, x_j) E[x_i] E[x_j] \\\\
        &+  Var[x_i] E[x_j]^2 + Var[x_j] E[x_i]^2

    Arguments
    ---------
    mean: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        1D array mean values.
    var: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        1D array variance values.
    cov: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        3D array covariance. Shape is ``(nblock, ndata, ndata)``.

    Returns
    -------
    new_cov: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        3D array propagated covariance.
    """
    nblock = cov.shape[0]
    ndata = cov.shape[1]
    new_cov = np.empty_like(cov)

    for jj in range(nblock):
        i1 = jj * ndata
        i2 = i1 + ndata
        v = var[i1:i2]
        m = mean[i1:i2]

        np.outer(v, v, out=new_cov[jj])
        new_cov[jj] += cov[jj]**2
        new_cov[jj] += 2 * cov[jj] * np.outer(m, m)

        C1 = np.outer(v, m**2)
        new_cov[jj] += C1 + C1.T

    return new_cov


def fft_gaussian_smooth(x, sigma_pix=20, mode='edge'):
    """ My Gaussian smoother using FFTs. Input array is padded with edge
    values at the boundary by default. Pad size is ``3*sigma_pix``.

    Arguments
    ---------
    x: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        1D array to smooth.
    sigma_pix: float, default: 20
        Smoothing Gaussian sigma in terms of number of pixels.
    mode: str
        Padding method. See :external+numpy:func:`numpy.pad` for options.

    Returns
    ---------
    y: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Smoothed x values. Same size as x.
    """
    # Pad the input array to get rid of annoying edge effects
    # Pad values are set to the edge value
    pad_size = max(1, int(3 * sigma_pix))
    arrsize = x.size + 2 * pad_size
    padded_arr = np.pad(x, pad_size, mode=mode)

    kvals = np.fft.rfftfreq(arrsize)
    smerror_k = np.fft.rfft(padded_arr) * np.exp(-(kvals * sigma_pix)**2 / 2.)

    y = np.fft.irfft(smerror_k, n=arrsize)[pad_size:-pad_size]

    return y


def get_median_outlier_mask(flux, ivar, esigma=3.5, nwindow=11):
    """Calculates median[i] of flux and sigma = 1.4826 * MAD[i] based on moving
    median filter of size ``nwindow``. Returns a boolean array to mask pixels
    if ``abs(f[i] - median[i]) > esigma * max(n[i], sigma[i])``.

    Arguments
    ---------
    flux: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Flux array.
    ivar: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Inverse variance array.
    esigma: float, default: 3.5
        Sigma to identify outliers via MAD.
    nwindow: int, default: 11
        Window size of the moving median filter.

    Returns
    ---------
    mask: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Bool array. Mask if true.
    """
    w = ivar != 0
    mask = np.zeros(flux.size, dtype=bool)
    mask[~w] = True

    if not any(w):
        return mask

    abs_diff = np.abs(flux[w] - median_filter(flux[w], nwindow))
    isig_mad = np.fmin(
        1.0, 1.0 / (1.4826 * median_filter(abs_diff, nwindow) + 1e-12))
    isig_cut = np.fmin(np.sqrt(ivar[w]), isig_mad) / esigma
    mask[w] = abs_diff * isig_cut > 1.0
    return mask


def get_smooth_ivar(ivar, sigma_pix=20, esigma=3.5):
    """ Smoothing ``ivar`` values to reduce signal-noise coupling.

    Smoothing is done on ``error=1/sqrt(ivar)``, while replacing ``ivar=0`` and
    outliers in ``error`` values with the median. These replaced values are put
    back in in the final result.

    Arguments
    ---------
    ivar: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Inverse variance array.
    sigma_pix: float, default: 20
        Smoothing Gaussian sigma in terms of number of pixels.
    esigma: float, default: 3.5
        Sigma to identify outliers via MAD.

    Returns
    ---------
    ivar2: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Smoothed ivar values. Outliers and masked values are put back in.
    """
    error = np.empty_like(ivar)
    w1 = ivar > 0
    error[w1] = 1 / np.sqrt(ivar[w1])
    median_err = np.median(error[w1])
    error[~w1] = median_err

    # Isolate high noise pixels
    mad = np.median(np.abs(error[w1] - median_err)) * 1.4826
    w2 = (error - median_err) > esigma * mad
    err_org = error[w2].copy()

    # Replace them with the median
    error[w2] = median_err
    error = fft_gaussian_smooth(error, sigma_pix)

    # Restore values of bad pixels
    error[w2] = err_org
    ivar2 = 1 / error**2
    ivar2[~w1] = 0

    return ivar2


class FastLinear1DInterp():
    """Fast interpolator class for equally spaced data. Out of domain points
    are linearly extrapolated without producing any warnings or errors.

    Uses :func:`_fast_eval_interp1d_lin`.

    Example::

        one_interp = FastLinear1DInterp(0., 1., np.ones(3))
        one_interp(5) # = 1

    Parameters
    ----------
    xp0: float
        Initial x point for interpolation data.
    dxp: float
        Spacing of x points.
    fp: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Function calculated at interpolation points.
    copy: bool, default: False
        Copy input data, specifically fp.
    ep: :external+numpy:py:class:`ndarray <numpy.ndarray>`, optional
        Error on fp points. Not used! Bookkeeping purposes only.
    """

    def __init__(self, xp0, dxp, fp, copy=False, ep=None):
        self.xp0 = float(xp0)
        self.dxp = float(dxp)
        if copy:
            self.fp = fp.copy()
        else:
            self.fp = fp
        self.ep = ep

    def __call__(self, x):
        return _fast_eval_interp1d_lin(x, self.xp0, self.dxp, self.fp)

    def reset(self, fp, copy=False, ep=None):
        if copy:
            self.fp = fp.copy()
        else:
            self.fp = fp

        self.ep = ep


class FastCubic1DInterp():
    """ Fast cubic spline for equally spaced data. Out of domain points
    are extrapolated without producing any warnings or errors.

    This class supports natural, not-a-knot boundary conditions. Uses
    :func:`_spline_cubic_natural`, :func:`_spline_cubic_notaknot`, and
    :func:`_fast_eval_interp1d_cubic`.

    Parameters
    ----------
    xp0: float
        Initial x point for interpolation data.
    dxp: float
        Spacing of x points.
    fp: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Function calculated at interpolation points.
    copy: bool, default: False
        Copy input data, specifically fp.
    ep: :external+numpy:py:class:`ndarray <numpy.ndarray>`, optional
        Error on fp points. Not used! Bookkeeping purposes only.
    bc_type: str, default: 'not-a-knot'
        Boundary condition type. Other option is 'natural'. See
        :external+scipy:py:class:`scipy.interpolate.CubicSpline`.
    """

    def __init__(
            self, xp0, dxp, fp, copy=False, ep=None, bc_type='not-a-knot'
    ):
        self.xp0 = float(xp0)
        self.dxp = float(dxp)
        if copy:
            self.fp = fp.copy()
        else:
            self.fp = fp
        self._bc_type = bc_type

        if self._bc_type == 'not-a-knot':
            self._y2p = _spline_cubic_notaknot(fp, dxp)
        elif self._bc_type == 'natural':
            self._y2p = _spline_cubic_natural(fp, dxp)
        else:
            raise Exception("Unknown bc_type in FastCubic1DInterp.")

        self.ep = ep

    def __call__(self, x):
        return _fast_eval_interp1d_cubic(
            x, self.xp0, self.dxp, self.fp, self._y2p)

    def reset(self, fp, copy=False, ep=None):
        if copy:
            self.fp = fp.copy()
        else:
            self.fp = fp

        self.ep = ep

        if self._bc_type == 'not-a-knot':
            self._y2p = _spline_cubic_notaknot(fp, self.dxp)
        elif self._bc_type == 'natural':
            self._y2p = _spline_cubic_natural(fp, self.dxp)


class SubsampleCov():
    """Utility class to store all subsamples with weights and calculate
    the delete-one Jackknife covariance matrix.

    Usage::

        subsampler = SubsampleCov(ndata, nsamples)

        for measurement, weights in SomeDataAfterFunction:
            # You can do this more than nsamples times
            subsampler.add_measurement(measurement, weights)

        mean, covariance = subsampler.get_mean_n_cov()

    .. warning::

            You cannot call :meth:`add_measurement` after calling
            :meth:`get_mean`, :meth:`get_mean_n_cov`,
            or :meth:`get_mean_n_var`.

    Parameters
    ----------
    ndata: int or tuple(int)
        Size or shape of the data vector. If tuple, it should be
        ``(nset, size1d)``. For example, 3 quantities share the same
        weights, data vector shape should pass ``ndata=(3, size1d)``.
    nsamples: int
        Number of samples. You can add more measurements then this.
    istart: int, default: 0
        Start index for the subsampling array

    Attributes
    ----------
    _istart: int
        Sampler initial index.
    _isample: int
        Sample counter. Wraps around nsamples
    _is_normalized: bool
        If the weights are normalized. Keeps track if :func:`_normalize` is
        called.
    all_measurements: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        3D array of shape ``(nsamples, nset, ndata)``.
    all_weights: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        3D array of shape ``(nsamples, nset, ndata)``.
    mean: :class:`ndarray <numpy.ndarray>` or None
        Mean. 2D array of shape ``(nset, ndata)``
    variance: :class:`ndarray <numpy.ndarray>` or None
        Variance. 2D array of shape ``(nset, ndata)``
    covariance: list(:class:`ndarray <numpy.ndarray>`) or None
        Covariance. 2D arrays of shape ``(ndata, ndata)`` or 3D arrays of
        shape ``(nblock, blockdim, blockdim)``.
    """

    def __init__(self, ndata, nsamples, istart=0):
        self.nsamples = nsamples
        self._istart = istart % nsamples
        self._isample = self._istart

        if isinstance(ndata, int):
            newshape = (nsamples, 1, ndata)
            self.ndata = ndata
        elif isinstance(ndata, tuple):
            newshape = (nsamples, ndata[0], ndata[1])
            self.ndata = ndata[1]
        else:
            raise QsonicException("ndata must be int or tuple of ints.")

        self.all_measurements = np.zeros(newshape)
        self.all_weights = np.zeros((nsamples, 1, self.ndata))
        self._is_normalized = False

        self.mean = None
        self.covariance = None
        self.variance = None
        self.xdiff = None

    def add_measurement(self, xvec, wvec):
        """ Adds a measurement to the sample.

        You can call this function more then ``nsamples`` times. The provided
        measurement should be weighted, but unnormalized, i.e. ``xvec=wi*xi``.
        After a mean or covariance is obtained, you cannot add
        more measurements.

        Arguments
        ---------
        xvec: :class:`ndarray <numpy.ndarray>`
            Data (measurement) vector.
        wvec: :class:`ndarray <numpy.ndarray>`
            1D weight vector.

        Raises
        ---------
        RuntimeError
            If the object is normalized by calling
            ``_normalize, get_mean, get_mean_n_cov`` and
            ``get_mean_n_var``.
        """
        if self._is_normalized:
            raise RuntimeError(
                "SubsampleCov has already been normalized. "
                "You cannot add more measurements.")

        self.all_measurements[self._isample, :, :] += xvec
        self.all_weights[self._isample, 0, :] += wvec
        self._isample = (self._isample + 1) % self.nsamples

    def allreduce(self, comm, inplace):
        """Sums statistics from all MPI process.

        .. note::

            Call this with ``inplace=MPI.IN_PLACE``.

        Arguments
        ---------
        comm: MPI.COMM_WORLD
            MPI comm object for Allreduce
        inplace: BufSpec
            MPI.IN_PLACE
        """

        comm.Allreduce(inplace, self.all_measurements)
        comm.Allreduce(inplace, self.all_weights)

    def _normalize(self):
        self.all_measurements = np.divide(
            self.all_measurements, self.all_weights,
            out=self.all_measurements, where=self.all_weights != 0)
        norm = self.all_weights.sum(0)
        self.all_weights = np.divide(
            self.all_weights, norm, out=self.all_weights, where=norm != 0)

        self._is_normalized = True

    def get_mean(self):
        """ Get the mean of all subsamples.

        .. warning::

            You cannot call :meth:`add_measurement` after calling this unless
            you :meth:`reset`.

        Returns
        -------
        mean: :class:`ndarray <numpy.ndarray>`
            Mean.
        """
        if not self._is_normalized:
            self._normalize()

        self.mean = np.einsum(
            'ijk,ik->jk', self.all_measurements, self.all_weights[:, 0, :])

        return self.mean

    def _get_xdiff(self, mean_xvec, bias_correct=False):
        # remove one measurement, then renormalize
        jack_i = (
            mean_xvec - self.all_measurements * self.all_weights
        ) / (1 - self.all_weights + np.finfo(float).eps)
        mean_jack = np.mean(jack_i, axis=0)

        if bias_correct:
            bias_jack = (self.nsamples - 1) * (mean_jack - mean_xvec)
            mean_xvec -= bias_jack

        xdiff = jack_i - mean_jack

        return mean_xvec, xdiff

    def _get_block_covariance(self, x, blockdim):
        nblock = self.ndata // blockdim
        cov = np.empty((nblock, blockdim, blockdim), dtype=np.float_)
        for kk in range(nblock):
            y = x[:, kk * blockdim:(kk + 1) * blockdim]
            cov[kk] = (y.T @ y) * (self.nsamples - 1) / self.nsamples

        return cov

    def get_mean_n_cov(self, indices=None, blockdim=None, bias_correct=False):
        """ Get the mean and covariance of the mean using delete-one Jackknife.

        Also sets :attr:`mean` and :attr:`covariance`.

        .. warning::

            You cannot call :meth:`add_measurement` after calling this unless
            you :meth:`reset`.

        Arguments
        ---------
        indices: list(int), default: None
            Data set indices to estimate the covariance.
        blockdim: int, default: None
            Calculate covariance by this block size instead of the full space.
        bias_correct: bool, default: False
            Jackknife bias correction term for the mean.

        Returns
        -------
        mean: :class:`ndarray <numpy.ndarray>`
            Mean.
        cov: list(:class:`ndarray <numpy.ndarray>`)
            Covariances of the mean.
        """
        if self.mean is None:
            mean_xvec = self.get_mean()
            self.mean, self.xdiff = self._get_xdiff(mean_xvec, bias_correct)

        if indices is None:
            indices = range(self.all_measurements.shape[1])

        if blockdim is not None:
            assert (self.ndata % blockdim == 0)

        self.covariance = [None] * self.all_measurements.shape[1]
        for jj in indices:
            x = self.xdiff[:, jj, :]

            if blockdim is None:
                cov = (x.T @ x) * (self.nsamples - 1) / self.nsamples
            else:
                cov = self._get_block_covariance(x, blockdim)
            self.covariance[jj] = cov

        return self.mean, self.covariance

    def get_mean_n_var(self, bias_correct=False):
        """ Get the mean and variance of themean (i.e. diagonal of the
        covariance) using delete-one Jackknife.

        .. warning::

            You cannot call :meth:`add_measurement` after calling this unless
            you :meth:`reset`.

        Arguments
        ---------
        bias_correct: bool, default: False
            Jackknife bias correction term for the mean.

        Returns
        -------
        mean_xvec: :class:`ndarray <numpy.ndarray>`
            Mean.
        var_xvec: :class:`ndarray <numpy.ndarray>`
            Variance of the mean. 1D array
        """
        if self.mean is None:
            mean_xvec = self.get_mean()
            self.mean, self.xdiff = self._get_xdiff(mean_xvec, bias_correct)

        self.variance = (np.einsum('ijk,ijk->jk', self.xdiff, self.xdiff)
                         * (self.nsamples - 1) / self.nsamples)

        return self.mean, self.variance

    def reset(self):
        self._isample = self._istart

        self.all_measurements.fill(0)
        self.all_weights.fill(0)
        self._is_normalized = False

        self.mean = None
        self.covariance = None
        self.variance = None
        self.xdiff = None



import logging
import warnings

import fitsio
import numpy as np

from healpy import ang2pix

from qsonic.mathtools import FastCubic1DInterp
from qsonic.continuum_models.base_continuum_model import BaseContinuumModel


def _find_append_continuum(
        specs, wave_rf_1, dwave_rf, common_targetids, continua
):
    for spec in specs:
        idx = np.nonzero(common_targetids == spec.targetid)[0]
        if idx.size == 0:
            continue

        idx = idx[0]
        spec.cont_params['valid'] = True
        spec.cont_params['input_w1'] = wave_rf_1
        spec.cont_params['input_dwave'] = dwave_rf
        spec.cont_params['input_data'] = continua[idx]


class InputContinuumModel(BaseContinuumModel):
    """Input continuum model class.

    Input continua should be in the rest frame with equal wavelength spacing.
    These continua are then interpolated using a cubic spline. Uses fiducials
    for mean flux and varlss interpolation.

    The files are assumed to be organized by healpix **nside=8** with
    **nested** ordering. For example, ``input-continuum-{nside}-{pixnum}.fits``
    , where
    ``pixnum = healpy.ang2pix(nside, ra_deg, dec_deg, lonlat=True, nest=True)``
    . Each file should have at least two extensions: **FIBERMAP** and
    **CONTINUA**.

    - **FIBERMAP** extension should have the typical DESI fibermap columns:
      ``TARGETID, Z, RA, DEC, PROGRAM, SURVEY``. Each ``TARGETID`` should occur
      once.

    - **CONTINUA** extention should be an ImageHDU and have the quasar continua
      in the same order of ``TARGETID`` s in **FIBERMAP**. Its header should
      declare the initial rest-frame wavelength by ``WAVE1`` key and the
      rest-frame wavelength step by ``DWAVE`` key. The data should be in
      ``(nqso, nwave)`` shape.

    Parameters
    ----------
    input_dir: str
        Input directory for all input-continuum FITS files.
    meanflux_interp: FastLinear1DInterp
        Interpolator for mean flux. If fiducial is not set, this equals to 1.
    varlss_interp: FastLinear1DInterp or FastCubic1DInterp
        Cubic spline for var_lss if fitting. Linear if from file.
    eta_interp: FastCubic1DInterp
        Interpolator for eta.
    nside: int, default: 8
        Healpy nside.
    """

    def __init__(
            self, input_dir, meanflux_interp, varlss_interp, eta_interp,
            nside=8
    ):
        self.input_dir = input_dir
        self.meanflux_interp = meanflux_interp
        self.varlss_interp = varlss_interp
        self.eta_interp = eta_interp
        self.nside = nside

    def _read_continua_one_file(self, pixnum, targetids):
        """Reads one input continuum file.

        Arguments
        ---------
        pixnum: int
            Healpix number
        targetids: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            TARGETIDs in this healpix

        Returns
        -------
        wave_rf_1: float
            First wavelength in the rest frame.
        dwave_rf: float
            Rest-frame wavelength step
        common_targetids: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            TARGETIDs that are found in the file.
        continua: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            2D array of shape ``(common_targetids.size, nwave)``

        Raises
        ------
        Exception
            If the number of quasars in **FIBERMAP** and **CONTINUA** extension
            do not match.
        RuntimeWarning
            If the file cannot be read. In this case, all objects are assumed
            to have invalid continuum.
        """
        fname = f"{self.input_dir}/input-continuum-{self.nside}-{pixnum}.fits"

        try:
            with fitsio.FITS(fname) as fitsfile:
                fbrmap = fitsfile['FIBERMAP'].read(columns='TARGETID')
                hdr = fitsfile['CONTINUA'].read_header()
                continua = fitsfile['CONTINUA'].read().astype(np.float64)
        except Exception as e:
            warnings.warn(
                f"InputContinuumModel cannot read file {fname}. "
                f"Reason {e}.")
            return 0, 0, None, None

        if fbrmap.size != continua.shape[0]:
            raise Exception(
                "InputContinuumModel file error. The number of quasars do not "
                f"match between FIBERMAP and CONTINUA in {fname}.")

        common_targetids, idx_fbr, _ = np.intersect1d(
            fbrmap, targetids, return_indices=True)
        return hdr['WAVE1'], hdr['DWAVE'], common_targetids, continua[idx_fbr]

    def _read_continua(self, spectra_list):
        """Reads and sets input continuum for all elements in ``spectra_list``.
        If a TARGETID is not found in its expected file, its continuum will be
        set to invalid.

        If found, :attr:`qsonic.spectrum.Spectrum.cont_params` dictionary gains
        the following::

            cont_params['valid'] = True
            cont_params['input_w1'] (float): First rest-frame wavelength
            cont_params['input_dwave'] (float): Rest-frame wavelength spacing
            cont_params['input_data'] (ndarray): Input continuum

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects.
        """
        local_catalog = np.hstack([_.catrow for _ in spectra_list]).copy()
        local_catalog['HPXPIXEL'] = ang2pix(
            self.nside, local_catalog['RA'], local_catalog['DEC'],
            lonlat=True, nest=True)

        # local_catalog is not sorted
        idx_sort = local_catalog.argsort(order=['HPXPIXEL', 'TARGETID'])
        spectra_list = np.array(spectra_list)[idx_sort]
        local_catalog = local_catalog[idx_sort]

        unique_pix, s = np.unique(local_catalog['HPXPIXEL'], return_index=True)
        hpx_split_catalog = np.split(local_catalog, s[1:])
        hpx_split_spectra_list = np.split(spectra_list, s[1:])

        for cat, specs in zip(hpx_split_catalog, hpx_split_spectra_list):
            pixnum = cat['HPXPIXEL'][0]
            targetids = cat['TARGETID']
            wave_rf_1, dwave_rf, common_targetids, continua = \
                self._read_continua_one_file(pixnum, targetids)

            if wave_rf_1 == 0:
                continue

            _find_append_continuum(
                specs, wave_rf_1, dwave_rf, common_targetids, continua)

    def fit_continuum(self, spec):
        """Input continuum reduction. Uses fiducials for mean flux and varlss
        interpolation. Continuum is interpolated using a cubic spline.

        Arguments
        ---------
        spec: Spectrum
            Spectrum object to add true continuum.
        """
        if not spec.cont_params['valid']:
            spec.cont_params['cont'] = None
            spec.cont_params['chi2'] = -1
            return

        input_cont_interp = FastCubic1DInterp(
            spec.cont_params['input_w1'], spec.cont_params['input_dwave'],
            spec.cont_params['input_data'])

        for arm, wave_arm in spec.forestwave.items():
            cont_est = input_cont_interp(wave_arm / (1 + spec.z_qso))

            if any(cont_est <= 0) or any(np.isnan(cont_est)):
                spec.cont_params['valid'] = False
                break

            cont_est *= self.meanflux_interp(wave_arm)
            spec.cont_params['cont'][arm] = cont_est

        spec.set_forest_weight(self.varlss_interp, self.eta_interp)
        spec.calc_continuum_chi2()

        if not spec.cont_params['valid']:
            spec.cont_params['cont'] = None
            spec.cont_params['chi2'] = -1

    def init_spectra(self, spectra_list):
        """ Reads input continuum. Initializes
        :attr:`cont_params <qsonic.spectrum.Spectrum.cont_params>` for a list
        of Spectrum objects.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects.
        """
        logging.info("Initializing input continuum.")

        for spec in spectra_list:
            spec.cont_params['method'] = 'input'
            spec.cont_params['valid'] = False
            spec.cont_params['x'] = np.zeros(1)
            spec.cont_params['xcov'] = np.eye(1)
            spec.cont_params['dof'] = spec.get_real_size()
            spec.cont_params['cont'] = {}

        self._read_continua(spectra_list)

    def stacks_residual_flux(self):
        """:meth:`stack_spectra` stacks actual continuum values.

        Returns
        -------
        False: bool
        """
        return False



import argparse
import logging
import time

from os import makedirs as os_makedirs

import numpy as np

from qsonic import QsonicException
import qsonic.calibration
import qsonic.catalog
import qsonic.io
import qsonic.spectrum
import qsonic.masks
from qsonic.mpi_utils import mpi_parse
from qsonic.picca_continuum import (
    PiccaContinuumFitter, add_picca_continuum_parser)


def get_parser(add_help=True):
    """Constructs the parser needed for the script.

    Arguments
    ---------
    add_help: bool, default: True
        Add help to parser.

    Returns
    -------
    parser: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(
        add_help=add_help,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser = qsonic.io.add_io_parser(parser)
    parser = qsonic.calibration.add_calibration_parser(parser)

    analysis_group = parser.add_argument_group('Analysis options')
    analysis_group.add_argument(
        "--smoothing-scale", default=16., type=float,
        help="Smoothing scale for pipeline noise in A.")
    analysis_group.add_argument(
        "--min-rsnr", type=float, default=0,
        help="Minium SNR <F/sigma> above Lya.")
    analysis_group.add_argument(
        "--min-forestsnr", type=float, default=0,
        help="Minium SNR <F/sigma> within the forest.")
    analysis_group.add_argument(
        "--skip", type=qsonic.io._float_range(0, 1), default=0.2,
        help="Skip short spectra lower than given ratio.")

    parser = qsonic.spectrum.add_wave_region_parser(parser)
    parser = qsonic.masks.add_mask_parser(parser)
    parser = add_picca_continuum_parser(parser)

    return parser


def args_logic_fnc_qsonic_fit(args):
    args.arms = list(set(args.arms))
    args.true_continuum |= (args.continuum_model == "true")

    condition_msg = [
        (args.true_continuum and not args.mock_analysis,
            "True continuum is only applicable to mock analysis."),
        (args.true_continuum and not args.fiducial_meanflux,
            "True continuum analysis requires fiducial mean flux."),
        (args.true_continuum and not args.fiducial_varlss,
            "True continuum analysis requires fiducial var_lss."),
        (args.true_continuum and args.input_continuum_dir,
            "Both true continuum and input continuum cannot be set."),
        (args.wave2 <= args.wave1,
            "wave2 must be greater than wave1."),
        (args.forest_w2 <= args.forest_w1,
            "forest_w2 must be greater than forest_w1."),
        (args.mock_analysis and args.tile_format,
            "Mock analysis in tile format is not supported."),
        (args.tile_format and args.save_by_hpx,
            "Cannot save deltas in healpixes in tile format."),
        (args.exposures != "disable" and args.tile_format,
            "Cannot save exposures in tile format.")
    ]

    condition_msg = [_ for _ in condition_msg if _[0]]

    for c, msg in condition_msg:
        logging.error(msg)

    if args.true_continuum:
        args.continuum_model = "true"
    if args.input_continuum_dir:
        args.continuum_model = "input"

    return not condition_msg


def mpi_read_spectra_local_queue(local_queue, args, comm):
    """ Read local spectra for the MPI rank. Set forest and observed wavelength
    range.

    Arguments
    ---------
    local_queue: list(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        Catalog from :func:`qsonic.catalog.mpi_get_local_queue`. Each
        element is a catalog for one healpix.
    args: argparse.Namespace
        Options passed to script.
    comm: MPI.COMM_WORLD
        Communication object for reducing data.

    Returns
    ---------
    spectra_list: list(Spectrum)
        Spectrum objects for the local MPI rank.
    """
    start_time = time.time()
    read_mode = "exposures" if args.exposures == "before" else "spectra"
    logging.info(f"Reading {read_mode}.")

    # skip_resomat = args.skip_resomat or not args.mock_analysis

    readerFunction = qsonic.io.get_spectra_reader_function(
        args.input_dir, args.arms, args.mock_analysis, args.skip_resomat,
        args.true_continuum, args.tile_format, args.exposures == "before")

    spectra_list = []
    # Each process reads its own list
    for cat in local_queue:
        local_specs = readerFunction(cat)

        for spec in local_specs:
            spec.set_forest_region(
                args.wave1, args.wave2, args.forest_w1, args.forest_w2)
            spec.remove_nonforest_pixels()

        spectra_list.extend([
            spec for spec in local_specs
            if spec.forestwave and spec.rsnr >= args.min_rsnr
        ])

    if args.coadd_arms == "before":
        logging.info("Coadding arms with pure IVAR weights.")
        for spec in spectra_list:
            spec.coadd_arms_forest()

    nspec_all = comm.reduce(len(spectra_list))
    etime = (time.time() - start_time) / 60  # min
    logging.info(f"All {nspec_all} {read_mode} are read in {etime:.1f} mins.")

    return spectra_list


def mpi_noise_flux_calibrate(spectra_list, args, comm, mpi_rank):
    if args.noise_calibration:
        logging.info("Applying noise calibration (eta only).")
        ncal = qsonic.calibration.NoiseCalibrator(
            args.noise_calibration, comm, mpi_rank,
            add_varlss=False)
        ncal.apply(spectra_list)

    if args.flux_calibration:
        logging.info("Applying flux calibration.")
        fcal = qsonic.calibration.FluxCalibrator(
            args.flux_calibration, comm, mpi_rank)
        fcal.apply(spectra_list)


def mpi_read_masks(local_queue, args, comm, mpi_rank):
    """ Read and set masking objects. Broadcast from the master process if
    necessary. See :mod:`qsonic.masks` for
    :class:`SkyMask <qsonic.masks.SkyMask>`,
    :class:`BALMask <qsonic.masks.BALMask>` and
    :class:`DLAMask <qsonic.masks.DLAMask>`.

    Arguments
    ---------
    local_queue: list(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        Catalog from :func:`qsonic.catalog.mpi_get_local_queue`.
    args: argparse.Namespace
        Options passed to script.
    comm: MPI.COMM_WORLD
        Communication object for broadcasting data.
    mpi_rank: int
        Rank of the MPI process

    Returns
    ---------
    maskers: list(Masks)
        Mask objects from `qsonic.masks`.
    """
    maskers = []

    if args.sky_mask:
        logging.info("Reading sky mask.")
        skymasker = qsonic.masks.SkyMask(args.sky_mask)

        maskers.append(skymasker)

    # BAL mask
    if args.bal_mask:
        logging.info("Checking BAL mask.")
        qsonic.masks.BALMask.check_catalog(local_queue[0])

        maskers.append(qsonic.masks.BALMask)

    # DLA mask
    if args.dla_mask:
        logging.info("Reading DLA mask.")
        local_targetids = np.concatenate(
            [cat['TARGETID'] for cat in local_queue])

        # Read catalog
        dlamasker = qsonic.masks.DLAMask(
            args.dla_mask, local_targetids, comm, mpi_rank,
            dla_mask_limit=0.8)

        maskers.append(dlamasker)

    return maskers


def apply_masks(maskers, spectra_list, mpi_rank=0):
    """ Apply masks in ``maskers`` to the local ``spectra_list``.

    See :mod:`qsonic.masks` for
    :class:`SkyMask <qsonic.masks.SkyMask>`,
    :class:`BALMask <qsonic.masks.BALMask>` and
    :class:`DLAMask <qsonic.masks.DLAMask>`. Masking is set by setting
    ``forestivar=0``. :class:`DLAMask <qsonic.masks.DLAMask>` further corrects
    for Lya and Lyb damping wings. Empty arms are removed after masking.

    Arguments
    ---------
    maskers: list(Masks)
        Mask objects from `qsonic.masks`.
    spectra_list: list(Spectrum)
        Spectrum objects for the local MPI rank.
    mpi_rank: int
        Rank of the MPI process
    """
    if not maskers:
        return

    start_time = time.time()
    logging.info("Applying masks.")
    for spec in spectra_list:
        for masker in maskers:
            masker.apply(spec)
        spec.drop_short_arms()
    etime = (time.time() - start_time) / 60   # min
    logging.info(f"Masks are applied in {etime:.1f} mins.")


def remove_short_spectra(spectra_list, lya1, lya2, skip_ratio):
    if not skip_ratio:
        return spectra_list

    logging.info("Removing short spectra.")
    dforest_wave = lya2 - lya1
    spectra_list = [spec for spec in spectra_list
                    if spec.is_long(dforest_wave, skip_ratio)]

    return spectra_list


def mpi_read_calibrate_mask_select_spectra(
        local_queue, maskers, args, comm, mpi_rank
):
    """ Read local spectra for the MPI rank. Set forest and observed wavelength
    range. Apply noise and flux calibration and maskers. Remove short spectra,
    and apply minimum forest SNR cut. Calls the following:

        - :func:`mpi_read_spectra_local_queue`,
        - :func:`mpi_noise_flux_calibrate`,
        - :func:`apply_masks`,
        - :func:`remove_short_spectra`.

    Arguments
    ---------
    local_queue: list(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        Catalog from :func:`qsonic.catalog.mpi_get_local_queue`. Each
        element is a catalog for one healpix.
    maskers: list(Masks)
        Mask objects from `qsonic.masks`.
    args: argparse.Namespace
        Options passed to script.
    comm: MPI.COMM_WORLD
        Communication object for reducing data.
    mpi_rank: int
        Rank of the MPI process

    Returns
    ---------
    spectra_list: list(Spectrum)
        Spectrum objects for the local MPI rank.
    """
    spectra_list = mpi_read_spectra_local_queue(local_queue, args, comm)

    mpi_noise_flux_calibrate(spectra_list, args, comm, mpi_rank)

    apply_masks(maskers, spectra_list, mpi_rank)

    # remove from sample if the number of pixels is small
    spectra_list = remove_short_spectra(
        spectra_list, args.forest_w1, args.forest_w2, args.skip)

    # Remove spectra with respect to forest snr
    spectra_list = [spec for spec in spectra_list
                    if spec.get_effective_meansnr() >= args.min_forestsnr]

    return spectra_list


def mpi_continuum_fitting(spectra_list, args, comm, mpi_rank):
    # Initialize continuum fitter & global functions
    logging.info("Initializing continuum fitter.")
    start_time = time.time()
    qcfit = PiccaContinuumFitter(args)

    # Fit continua
    # Stack all spectra in each process
    # Broadcast and recalculate global functions
    # Iterate
    logging.info("Fitting continuum.")
    qcfit.iterate(spectra_list)

    valid_spectra = list(qsonic.spectrum.valid_spectra(spectra_list))

    # Final noise calibration for additive var_lss.
    if args.noise_calibration and args.varlss_as_additive_noise:
        logging.info("Applying noise calibration (varlss only).")
        ncal = qsonic.calibration.NoiseCalibrator(
            args.noise_calibration, comm, mpi_rank,
            add_varlss=True, no_eta=True)
        ncal.apply(valid_spectra)

        for spec in valid_spectra:
            spec.set_smooth_forestivar(spec._smoothing_scale)
            spec.set_forest_weight(qcfit.varlss_interp, qcfit.eta_interp)

    if args.coadd_arms == "after":
        logging.info("Coadding arms.")
        for spec in valid_spectra:
            spec.coadd_arms_forest(qcfit.varlss_interp, qcfit.eta_interp)
            spec.calc_continuum_chi2()

    qcfit.save_contchi2_catalog(spectra_list)

    # Keep only valid spectra
    spectra_list = valid_spectra

    etime = (time.time() - start_time) / 60  # min
    logging.info(f"Continuum fitting and tweaking took {etime:.1f} mins.")

    return spectra_list


def mpi_read_exposures_after(spectra_list, maskers, args, comm, mpi_rank):
    """Creates a local catalog from spectra_list and reads exposures. Coadding
    of arms is always done with IVAR only as weights (exposures are not coadded
    ). RSNR and forest SNR cuts are still applied. CONT is copied from the
    exposure coadded spectra. The extension name in the delta files will be
    ``TARGETID_ARM_EXPID``.

    Arguments
    ---------
    spectra_list: list(Spectrum)
        Spectrum objects for the local MPI rank.
    maskers: list(Masks)
        Mask objects from `qsonic.masks`.
    args: argparse.Namespace
        Options passed to script.
    comm: MPI.COMM_WORLD
        Communication object for reducing data.
    mpi_rank: int
        Rank of the MPI process

    Returns
    ---------
    exposure_spectra_list: list(Spectrum)
        Spectrum objects for the local MPI rank. Note there will be duplicate
        TARGETIDs.
    """
    # Initialize continuum fitter & global functions
    logging.info("Reading exposures and applying RSNR cut.")
    start_time = time.time()

    local_catalog = np.hstack([_.catrow for _ in spectra_list])

    # local_catalog is not sorted
    idx_sort = local_catalog.argsort(order=['HPXPIXEL', 'TARGETID'])
    spectra_list = np.array(spectra_list)[idx_sort]
    local_catalog = local_catalog[idx_sort]
    # assert all(np.arange(idx_sort.size) == idx_sort)
    del idx_sort

    # exposures cannot be saved in tile format
    unique_pix, s = np.unique(local_catalog['HPXPIXEL'], return_index=True)
    local_queue = np.split(local_catalog, s[1:])
    del s, unique_pix

    args.exposures = "before"
    if args.coadd_arms != "disable":
        args.coadd_arms = "before"

    exposure_spectra_list = mpi_read_calibrate_mask_select_spectra(
        local_queue, maskers, args, comm, mpi_rank
    )

    # Match & assign cont, forest weight
    for spec in exposure_spectra_list:
        spec_coadd = spectra_list[
            np.nonzero(local_catalog['TARGETID'] == spec.targetid)[0][0]
        ]
        spec.cont_params = spec_coadd.cont_params.copy()
        spec.cont_params['cont'] = {}

        for arm, wave_arm in spec.forestwave.items():
            wave_coadd = spec_coadd.forestwave
            if arm not in wave_coadd:
                spec.drop_arm(arm)
                continue

            i1, i2 = 0, wave_arm.size
            j1 = int(round((wave_arm[0] - wave_coadd[arm][0]) / spec.dwave))
            if j1 < 0:
                i1 = -j1
                j1 = 0

            j2 = int(round((wave_coadd[arm][-1] - wave_arm[-1]) / spec.dwave))
            if j2 < 0:
                i2 += j2
                j2 = 0
            j2 = wave_coadd[arm].size - j2
            spec.slice(arm, i1, i2)

            spec.cont_params['cont'][arm] = \
                spec_coadd.cont_params['cont'][arm][j1:j2].copy()

            if spec.forestwave[arm].size != spec.cont_params['cont'][arm].size:
                spec.cont_params['valid'] = False

    etime = (time.time() - start_time) / 60  # min
    logging.info(
        "Reading, calibrating, masking and matching the continua "
        f"took {etime:.1f} mins.")

    return exposure_spectra_list


def mpi_run_all(comm, mpi_rank, mpi_size):
    args = mpi_parse(get_parser(), comm, mpi_rank,
                     args_logic_fnc=args_logic_fnc_qsonic_fit)

    if mpi_rank == 0 and args.outdir:
        os_makedirs(args.outdir, exist_ok=True)

    tol = (args.forest_w2 - args.forest_w1) * args.skip
    zmin_qso = args.wave1 / (args.forest_w2 - tol) - 1
    zmax_qso = args.wave2 / (args.forest_w1 + tol) - 1

    # read catalog
    local_queue = qsonic.catalog.mpi_get_local_queue(
        args.catalog, comm, mpi_rank, mpi_size, args.mock_analysis,
        args.tile_format, args.keep_surveys, zmin_qso, zmax_qso)

    # Blinding
    if args.mock_analysis:
        maxlastnight = None
    else:
        maxlastnight = np.max([_['LASTNIGHT'].max() for _ in local_queue])
        maxlastnight = comm.allreduce(maxlastnight, max)
    qsonic.spectrum.Spectrum.set_blinding(maxlastnight, args)

    # Read masks before data
    maskers = mpi_read_masks(local_queue, args, comm, mpi_rank)

    spectra_list = mpi_read_calibrate_mask_select_spectra(
        local_queue, maskers, args, comm, mpi_rank
    )

    # Create smoothed ivar as intermediate variable
    if args.smoothing_scale > 0:
        for spec in spectra_list:
            spec.set_smooth_forestivar(args.smoothing_scale)

    # Continuum fitting
    spectra_list = mpi_continuum_fitting(spectra_list, args, comm, mpi_rank)

    if args.exposures == "after":
        spectra_list = mpi_read_exposures_after(
            spectra_list, maskers, args, comm, mpi_rank)

    # Final cleaning. Especially important if not coadding arms.
    for spec in spectra_list:
        spec.drop_short_arms(args.forest_w1, args.forest_w2, args.skip)

    # Save deltas
    logging.info("Saving deltas.")
    qsonic.io.save_deltas(
        spectra_list, args.outdir,
        save_by_hpx=args.save_by_hpx, mpi_rank=mpi_rank)


def main():
    start_time = time.time()

    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    mpi_rank = comm.Get_rank()
    mpi_size = comm.Get_size()

    logging.basicConfig(
        format='%(asctime)s - %(levelname)s: %(message)s',
        datefmt='%Y/%m/%d %I:%M:%S %p',
        level=logging.DEBUG if mpi_rank == 0 else logging.CRITICAL)

    try:
        mpi_run_all(comm, mpi_rank, mpi_size)
    except QsonicException as e:
        logging.exception(e)
        exit(1)
    except Exception as e:
        logging.critical(
            f"Unexpected error on Rank{mpi_rank}: {e}. Abort.", exc_info=True)
        comm.Abort()

    etime = (time.time() - start_time) / 60  # min
    logging.info(f"Total time spent is {etime:.1f} mins.")



""" Masking module with classes for sky, BAL and DLA."""
import argparse

from astropy.io.ascii import read as asread
import numpy as np
from numpy.lib.recfunctions import rename_fields
import fitsio

from qsonic import QsonicException
from qsonic.mpi_utils import mpi_fnc_bcast


LIGHT_SPEED = 299792.458
"""float: Speed of light in km/s."""
sqrt_pi = 1.77245385091
"""float: Square root of pi."""
sqrt_2 = 1.41421356237
"""float: Square root of 2."""


def add_mask_parser(parser=None):
    """ Adds masking related arguments to parser. These arguments are grouped
    under 'Masking options'.

    Arguments
    ---------
    parser: argparse.ArgumentParser, default: None

    Returns
    ---------
    parser: argparse.ArgumentParser
    """
    if parser is None:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    mask_group = parser.add_argument_group('Masking options')

    mask_group.add_argument(
        "--sky-mask",
        help="Sky mask file.")
    mask_group.add_argument(
        "--bal-mask", action="store_true",
        help="Mask BALs (assumes it is in catalog).")
    mask_group.add_argument(
        "--dla-mask",
        help="DLA catalog to mask.")

    return parser


class SkyMask():
    """ Sky line masking object.

    Parameters
    ----------
    fname: str
        Filename to read by `astropy.io.ascii`. Must have four columns ordered
        as type, minimum wavelength, maximum wavelength and frame (must be 'RF'
        or 'OBS').
    comm: None or MPI.COMM_WORLD, default: None
    mpi_rank: int, default: 0
    """
    column_names = ('type', 'wave_min', 'wave_max', 'frame')
    """tuple: Assumed ordering of columns in the text file."""

    def __init__(self, fname, comm=None, mpi_rank=0):
        mask = mpi_fnc_bcast(
            asread, comm, mpi_rank,
            f"Error loading SkyMask from mask file {fname}.",
            fname, names=SkyMask.column_names)

        self.mask_rest_frame = mask[mask['frame'] == 'RF']
        self.mask_obs_frame = mask[mask['frame'] == 'OBS']

    def apply(self, spec):
        """ Apply the mask by setting **only** ``spec.forestivar`` and
        ``spec.forestflux`` to zero.

        Arguments
        ----------
        spec: Spectrum
            Spectrum object to mask.
        """
        for arm, wave_arm in spec.forestwave.items():
            w = np.zeros(wave_arm.size, dtype=bool)

            m1 = np.searchsorted(
                wave_arm,
                [self.mask_obs_frame['wave_min'],
                 self.mask_obs_frame['wave_max']]
            ).T

            m2 = np.searchsorted(
                wave_arm / (1.0 + spec.z_qso),
                [self.mask_rest_frame['wave_min'],
                 self.mask_rest_frame['wave_max']]
            ).T

            mask_idx_ranges = np.concatenate((m1, m2))
            for idx1, idx2 in mask_idx_ranges:
                w[idx1:idx2] = 1

            spec.forestflux[arm][w] = 0
            spec.forestivar[arm][w] = 0


class BALMask():
    """ BAL masking object.

    Does not need construction. Assumes BAL related columns are present in the
    catalog.
    """
    lines = np.array([
        ("lCIV", 1549),
        ("lSiIV2", 1403),
        ("lSiIV1", 1394),
        ("lNV", 1240.81),
        ("lLya", 1216.1),
        ("lCIII", 1175),
        ("lPV2", 1128),
        ("lPV1", 1117),
        ("lSIV2", 1074),
        ("lSIV1", 1062),
        ("lOIV", 1031),
        ("lOVI", 1037),
        ("lOI", 1039),
        ("lLyb", 1025.7),
        ("lLy3", 972.5),
        ("lCIII", 977.0),
        ("lNIII", 989.9),
        ("lLy4", 949.7)],
        dtype=[("name", "U10"), ("value", 'f8')])
    """:external+numpy:py:class:`ndarray <numpy.ndarray>`: Ion transition
    wavelengths in A."""

    expected_columns = [
        'VMIN_CIV_450', 'VMAX_CIV_450',
        'VMIN_CIV_2000', 'VMAX_CIV_2000'
    ]
    """list(str): Columns needed in catalog to mask wavelength ranges."""

    @staticmethod
    def check_catalog(catalog):
        """ Asserts if the required columns are present in the catalog.

        Arguments
        ----------
        catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        """
        if not all(col in catalog.dtype.names
                   for col in BALMask.expected_columns):
            raise QsonicException("Input catalog is missing BAL columns.")

    @staticmethod
    def apply(spec):
        """ Apply the mask by setting **only** ``spec.forestivar`` and
        ``spec.forestflux`` to zero.

        Arguments
        ----------
        spec: Spectrum
            Spectrum object to mask.
        """
        min_velocities = np.concatenate(
            (spec.catrow['VMIN_CIV_450'], spec.catrow['VMIN_CIV_2000']))
        max_velocities = np.concatenate(
            (spec.catrow['VMAX_CIV_450'], spec.catrow['VMAX_CIV_2000']))
        w = (min_velocities > 0) & (max_velocities > 0)
        min_velocities = min_velocities[w]
        max_velocities = max_velocities[w]
        num_velocities = min_velocities.size

        if num_velocities == 0:
            return

        bal_obs_lines = BALMask.lines['value'] * (1 + spec.z_qso)
        min_velocities = 1 - min_velocities / LIGHT_SPEED
        max_velocities = 1 - max_velocities / LIGHT_SPEED

        mask = np.empty(num_velocities * BALMask.lines.size,
                        dtype=[('wave_min', 'f8'), ('wave_max', 'f8')])
        mask['wave_min'] = np.outer(bal_obs_lines, max_velocities).ravel()
        mask['wave_max'] = np.outer(bal_obs_lines, min_velocities).ravel()

        for arm, wave_arm in spec.forestwave.items():
            w = np.zeros(wave_arm.size, dtype=bool)

            mask_idx_ranges = np.searchsorted(
                wave_arm,
                [mask['wave_min'], mask['wave_max']]
            ).T
            # Make sure first index comes before the second
            mask_idx_ranges.sort(axis=1)
            for idx1, idx2 in mask_idx_ranges:
                w[idx1:idx2] = 1

            spec.forestflux[arm][w] = 0
            spec.forestivar[arm][w] = 0


class DLAMask():
    """ DLA masking object.

    Maximum numbers for oscillator strengths and Einstein coefficients are
    picked from NIST.

    Parameters
    ----------
    fname: str
        FITS filename to read.
    local_targetids: :class:`ndarray <numpy.ndarray>` or None, default: None
        Remove DLAs if they are present in these TARGETIDs.
    comm: None or MPI.COMM_WORLD, default: None
    mpi_rank: int, default: 0
    dla_mask_limit: float, default: 0.8
    """
    qe = 4.803204e-10
    """float: Charge of electron in statC (cm^3/2 g^1/2 s^-1)."""
    me = 9.109384e-28
    """float: Mass of electron in g."""
    c_cms = LIGHT_SPEED * 1e5
    """float: Speed of light in cm/s."""
    aij_coeff = np.pi * qe**2 / me / c_cms
    """float: Derived quantity for transition strength in cm^2 s^-1."""

    wave_lya_A = 1215.67
    """float: Lya wavelength in A."""

    # I suppose we are to pick maximum for each from NIST?
    f12_lya = 4.1641e-01
    """float: Lya oscillator strength."""
    A12_lya = 6.2648e+08
    """float: Lya Einstein coefficient."""

    wave_lyb_A = 1025.7220
    """float: Lyb wavelength in A."""
    f12_lyb = 7.9142e-02
    """float: Lyb oscillator strength."""
    A12_lyb = 1.6725e+08
    """float: Lyb Einstein coefficient."""

    accepted_zcolnames = ["Z_DLA", "Z"]
    """list(str): Column names for the DLA redshift."""

    @staticmethod
    def H_tepper_garcia(a, u):
        """ Tepper-Garcia H function.

        Arguments
        ---------
        a: float
            Gamma / sigma.
        u: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            (wave / lambda_12 - 1) / sigma.

        Returns
        -------
        :external+numpy:py:class:`ndarray <numpy.ndarray>`
        """
        P = u**2 + 1e-12
        Q = 1.5 / P
        R = np.exp(-P)
        corr = (R**2 * (4 * P**2 + 7 * P + 4 + Q) - Q - 1)
        corr *= a / P / sqrt_pi
        return R - corr

    @staticmethod
    def voigt_tepper_garcia(x, sigma, gamma):
        """Tepper-Garcia voight function.

        Arguments
        ---------
        x: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            wave / lambda_12 - 1.
        sigma: float
            b / c.
        gamma: float

        Returns
        -------
        :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Voight profile.
        """
        a = gamma / sigma
        u = x / sigma
        return DLAMask.H_tepper_garcia(a, u)

    @staticmethod
    def get_optical_depth(wave_A, lambda12_A, log10N, b, f12, A12):
        """Optical depth for a transition line.

        Arguments
        ---------
        wave_A: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Wavelength array in A.
        lambda12_A: float
            Transition wavelength in A.
        log10N: float
            Log10 of column density.
        b: float
            Doppler parameter in km/s.
        f12: float
            Oscillator strength.
        A12: float
            Einstein coefficient.

        Returns
        -------
        tau: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Optical depth.
        """
        a12 = DLAMask.aij_coeff * f12
        gamma = A12 * lambda12_A * 1e-8 / 4 / np.pi / DLAMask.c_cms

        sigma_gauss = b / LIGHT_SPEED

        tau = DLAMask.voigt_tepper_garcia(
            wave_A / lambda12_A - 1, sigma_gauss, gamma)
        tau *= a12 * 10**(log10N - 13) * (lambda12_A / b) / sqrt_pi

        return tau

    @staticmethod
    def get_dla_flux(wave, z_dla, nhi, b=10.):
        """Normalized flux from Lya and Lyb wings of DLA.

        Arguments
        ---------
        wave: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Wavelength array in A.
        z_dla: float
            DLA redshift
        nhi: float
            Log10 of DLA column density.
        b: float, default: 10
            Doppler parameter in km/s.

        Returns
        -------
        F: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Normalized flux for DLA.
        """
        wave_rf = wave / (1 + z_dla)
        tau = DLAMask.get_optical_depth(
            wave_rf, DLAMask.wave_lya_A, nhi, b,
            DLAMask.f12_lya, DLAMask.A12_lya)
        tau += DLAMask.get_optical_depth(
            wave_rf, DLAMask.wave_lyb_A, nhi, b,
            DLAMask.f12_lyb, DLAMask.A12_lyb)
        return np.exp(-tau)

    @staticmethod
    def get_all_dlas(wave, spec_dlas):
        """Normalized flux from all DLAs in a sightline.

        Arguments
        ---------
        wave: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Wavelength array in A.
        spec_dlas: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Named ndarray with 'Z_DLA' and 'NHI'.

        Returns
        -------
        F: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Normalized flux.
        """
        transmission = np.ones(wave.size)
        for z_dla, nhi in spec_dlas[['Z_DLA', 'NHI']]:
            transmission *= DLAMask.get_dla_flux(wave, z_dla, nhi)

        return transmission

    @staticmethod
    def _read_catalog(fname):
        """ Read and return DLA catalog.

        Should be run on master. If error occurs, returns None.

        Arguments
        ---------
        fname: str
            FITS filename to read.

        Returns
        -------
        catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            DLA catalog.
        """
        z_colname = DLAMask.accepted_zcolnames[0]
        fts = fitsio.FITS(fname)

        fts_colnames = set(fts["DLACAT"].get_colnames())
        z_colname = fts_colnames.intersection(DLAMask.accepted_zcolnames)

        if not z_colname:
            fts.close()
            raise ValueError(
                "DLA mask error::Z colname has to be one of "
                f"{', '.join(DLAMask.accepted_zcolnames)}")

        z_colname = z_colname.pop()
        columns_list = ["TARGETID", z_colname, "NHI"]
        catalog = fts['DLACAT'].read(columns=columns_list)

        if z_colname != 'Z_DLA':
            catalog = rename_fields(catalog, {z_colname: 'Z_DLA'})

        fts.close()

        return catalog

    def __init__(
            self, fname, local_targetids=None, comm=None, mpi_rank=0,
            dla_mask_limit=0.8
    ):
        self.dla_mask_limit = dla_mask_limit

        catalog = mpi_fnc_bcast(
            DLAMask._read_catalog, comm, mpi_rank,
            f"Error loading DLAMask from file {fname}.",
            fname)

        if local_targetids is not None:
            w = np.isin(catalog['TARGETID'], local_targetids)
            catalog = catalog[w]
        catalog.sort(order='TARGETID')

        # Group DLA catalog into targetids
        self.unique_targetids, s = np.unique(
            catalog['TARGETID'], return_index=True)
        self.split_catalog = np.split(catalog, s[1:])

    def apply(self, spec):
        """ Apply the mask by setting ``spec.forestivar`` and
        ``spec.forestflux`` to zero.

        Arguments
        ----------
        spec: Spectrum
            Spectrum object to mask.
        """
        w = np.nonzero(self.unique_targetids == spec.targetid)[0]
        if w.size == 0:
            return

        idx = w[0]
        spec_dlas = self.split_catalog[idx]
        for arm, wave_arm in spec.forestwave.items():
            transmission = DLAMask.get_all_dlas(wave_arm, spec_dlas)
            # Turn off DLA correction for l_rf > l_lya
            transmission[np.searchsorted(
                wave_arm, (1 + spec.z_qso) * DLAMask.wave_lya_A):
            ] = 1
            w = transmission < self.dla_mask_limit
            transmission[w] = 1

            spec.forestivar[arm][w] = 0
            spec.forestflux[arm][w] = 0
            spec.forestflux[arm] /= transmission
            spec.forestivar[arm] *= transmission**2
